from __future__ import annotations

import logging
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Optional

if TYPE_CHECKING:
    from ....core.flows import FlowController, FlowDefinition
    from . import FlowRoutesState

_logger = logging.getLogger(__name__)


def build_flow_definition(
    repo_root: Path, flow_type: str, state: "FlowRoutesState"
) -> FlowDefinition:
    repo_root = repo_root.resolve()
    key = (repo_root, flow_type)
    with state.lock:
        if key in state.definition_cache:
            return state.definition_cache[key]

    from ....core.config import load_repo_config
    from ....core.runtime import RuntimeContext
    from ....flows.ticket_flow import build_ticket_flow_definition
    from ....integrations.agents.build_agent_pool import build_agent_pool

    if flow_type == "ticket_flow":
        config = load_repo_config(repo_root)
        engine = RuntimeContext(
            repo_root=repo_root,
            config=config,
        )
        agent_pool = build_agent_pool(engine.config)
        definition = build_ticket_flow_definition(
            agent_pool=agent_pool,
            auto_commit_default=engine.config.git_auto_commit,
            include_previous_ticket_context_default=(
                engine.config.ticket_flow.include_previous_ticket_context
            ),
        )
    else:
        from fastapi import HTTPException

        raise HTTPException(status_code=404, detail=f"Unknown flow type: {flow_type}")

    with state.lock:
        state.definition_cache[key] = definition
    return definition


def get_flow_controller(
    repo_root: Path, flow_type: str, state: "FlowRoutesState"
) -> FlowController:
    repo_root = repo_root.resolve()
    key = (repo_root, flow_type)
    with state.lock:
        if key in state.controller_cache:
            controller = state.controller_cache[key]
            try:
                controller.initialize()
            except Exception:
                pass
            return controller

    definition = build_flow_definition(repo_root, flow_type, state)
    controller = definition.build_controller(repo_root)
    with state.lock:
        state.controller_cache[key] = controller
    return controller


def get_flow_record(
    repo_root: Path, run_id: str, state: "FlowRoutesState"
) -> Optional[Dict[str, Any]]:
    from ...services import flow_store as flow_store_service
    from .runtime_service import recover_flow_store_if_possible

    try:
        store = flow_store_service.require_flow_store(repo_root, logger=_logger)
        if store is None:
            return None
        record = store.get_run(run_id)
        if record is None:
            return None
        return {
            "id": record.id,
            "flow_type": record.flow_type,
            "status": (
                record.status.value
                if hasattr(record.status, "value")
                else str(record.status)
            ),
            "current_ticket_index": record.current_ticket_index,
            "current_ticket_path": (
                str(record.current_ticket_path) if record.current_ticket_path else None
            ),
            "created_at": (
                record.created_at.isoformat()
                if hasattr(record.created_at, "isoformat")
                else str(record.created_at)
            ),
            "updated_at": (
                record.updated_at.isoformat()
                if hasattr(record.updated_at, "isoformat")
                else str(record.updated_at)
            ),
        }
    except Exception as exc:
        recovered = recover_flow_store_if_possible(repo_root, "ticket_flow", state, exc)
        if not recovered:
            return None
        try:
            store = flow_store_service.require_flow_store(repo_root, logger=_logger)
            if store is None:
                return None
            record = store.get_run(run_id)
            if record is None:
                return None
            return {
                "id": record.id,
                "flow_type": record.flow_type,
                "status": (
                    record.status.value
                    if hasattr(record.status, "value")
                    else str(record.status)
                ),
                "current_ticket_index": record.current_ticket_index,
                "current_ticket_path": (
                    str(record.current_ticket_path)
                    if record.current_ticket_path
                    else None
                ),
                "created_at": (
                    record.created_at.isoformat()
                    if hasattr(record.created_at, "isoformat")
                    else str(record.created_at)
                ),
                "updated_at": (
                    record.updated_at.isoformat()
                    if hasattr(record.updated_at, "isoformat")
                    else str(record.updated_at)
                ),
            }
        except Exception:
            return None



from __future__ import annotations

from typing import Any, Optional


def _select_context_usage_bucket(
    token_usage: Optional[dict[str, Any]],
) -> Optional[dict[str, Any]]:
    if not isinstance(token_usage, dict):
        return None
    last = token_usage.get("last")
    if isinstance(last, dict):
        return last
    total = token_usage.get("total")
    if isinstance(total, dict):
        return total
    return None


def _extract_context_usage_percent(
    token_usage: Optional[dict[str, Any]],
) -> Optional[int]:
    usage = _select_context_usage_bucket(token_usage)
    if not isinstance(usage, dict):
        return None
    total_tokens = usage.get("totalTokens")
    if not isinstance(token_usage, dict):
        return None
    context_window = token_usage.get("modelContextWindow")
    if not isinstance(total_tokens, int) or not isinstance(context_window, int):
        return None
    if context_window <= 0:
        return None
    percent_used = round(total_tokens / context_window * 100)
    percent_remaining = max(0, 100 - percent_used)
    return min(percent_remaining, 100)


def _format_tui_token_usage(token_usage: Optional[dict[str, Any]]) -> Optional[str]:
    if not isinstance(token_usage, dict):
        return None
    usage = _select_context_usage_bucket(token_usage)
    if not isinstance(usage, dict):
        return None
    total_tokens = usage.get("totalTokens")
    input_tokens = usage.get("inputTokens")
    output_tokens = usage.get("outputTokens")
    if not isinstance(total_tokens, int):
        return None
    parts = [f"Token usage: total {total_tokens}"]
    if isinstance(input_tokens, int):
        parts.append(f"input {input_tokens}")
    if isinstance(output_tokens, int):
        parts.append(f"output {output_tokens}")
    percent = _extract_context_usage_percent(token_usage)
    if percent is not None:
        parts.append(f"ctx {percent}%")
    return " ".join(parts)


def _format_turn_metrics(
    token_usage: Optional[dict[str, Any]],
    elapsed_seconds: Optional[float],
) -> Optional[str]:
    lines: list[str] = []
    if elapsed_seconds is not None:
        lines.append(f"Turn time: {elapsed_seconds:.1f}s")
    token_line = _format_tui_token_usage(token_usage)
    if token_line:
        lines.append(token_line)
    if not lines:
        return None
    return "\n".join(lines)



from __future__ import annotations

import hashlib
import logging
from pathlib import Path
from typing import Any, Optional

from ..manifest import load_manifest
from .chat_bindings import (
    DISCORD_STATE_FILE_DEFAULT,
    TELEGRAM_STATE_FILE_DEFAULT,
    preferred_non_pma_chat_notification_source_for_workspace,
)
from .config import load_hub_config
from .time_utils import now_iso
from .utils import canonicalize_path

logger = logging.getLogger(__name__)

_DISCORD_MESSAGE_MAX_LEN = 1900


def _normalize_optional_text(value: Any) -> Optional[str]:
    if value is None:
        return None
    text = value if isinstance(value, str) else str(value)
    text = text.strip()
    return text or None


def _resolve_state_path(
    hub_root: Path, raw_config: dict[str, Any], *, section: str, default_state_file: str
) -> Path:
    section_cfg = raw_config.get(section) if isinstance(raw_config, dict) else {}
    if not isinstance(section_cfg, dict):
        section_cfg = {}
    state_file = section_cfg.get("state_file")
    if not isinstance(state_file, str) or not state_file.strip():
        state_file = default_state_file
    state_path = Path(state_file)
    if not state_path.is_absolute():
        state_path = hub_root / state_path
    return state_path.resolve()


def _binding_matches_workspace(binding_workspace: Any, workspace_root: Path) -> bool:
    if not isinstance(binding_workspace, str) or not binding_workspace.strip():
        return False
    try:
        return canonicalize_path(Path(binding_workspace)) == canonicalize_path(
            workspace_root
        )
    except Exception:
        return False


def _repo_id_by_workspace_path(hub_root: Path) -> dict[str, str]:
    manifest_path = hub_root / ".codex-autorunner" / "manifest.yml"
    if not manifest_path.exists():
        return {}
    try:
        loaded = load_manifest(manifest_path, hub_root)
    except Exception:
        return {}
    mapping: dict[str, str] = {}
    for repo in loaded.repos:
        try:
            workspace_path = canonicalize_path(hub_root / repo.path)
        except Exception:
            continue
        mapping[str(workspace_path)] = repo.id
    return mapping


def _resolve_repo_id(
    *,
    repo_id: Any,
    workspace_path: Any,
    repo_id_by_workspace: dict[str, str],
) -> Optional[str]:
    normalized_repo_id = _normalize_optional_text(repo_id)
    if normalized_repo_id:
        return normalized_repo_id
    if not isinstance(workspace_path, str) or not workspace_path.strip():
        return None
    try:
        return repo_id_by_workspace.get(str(canonicalize_path(Path(workspace_path))))
    except Exception:
        return None


async def notify_preferred_bound_chat_for_workspace(
    *,
    hub_root: Path,
    workspace_root: Path,
    repo_id: Optional[str],
    message: str,
    correlation_id: str,
) -> dict[str, int]:
    text = str(message or "").strip()
    if not text:
        return {"targets": 0, "published": 0}

    try:
        raw_config = load_hub_config(hub_root).raw
    except Exception:
        raw_config = {}
    repo_id_by_workspace = _repo_id_by_workspace_path(hub_root)
    preferred_source = preferred_non_pma_chat_notification_source_for_workspace(
        hub_root=hub_root,
        raw_config=raw_config,
        workspace_root=workspace_root,
    )
    if preferred_source not in {"discord", "telegram"}:
        return {"targets": 0, "published": 0}

    created_at = now_iso()
    normalized_repo_id = _normalize_optional_text(repo_id)

    async def _notify_discord() -> dict[str, int]:
        from ..integrations.discord.rendering import (
            chunk_discord_message,
            format_discord_message,
        )
        from ..integrations.discord.state import DiscordStateStore
        from ..integrations.discord.state import OutboxRecord as DiscordOutboxRecord

        targets = 0
        published = 0
        store = DiscordStateStore(
            _resolve_state_path(
                hub_root,
                raw_config,
                section="discord_bot",
                default_state_file=DISCORD_STATE_FILE_DEFAULT,
            )
        )
        try:
            bindings = await store.list_bindings()
            channels: list[str] = []
            for binding in bindings:
                if bool(binding.get("pma_enabled")):
                    continue
                if not _binding_matches_workspace(
                    binding.get("workspace_path"), workspace_root
                ):
                    continue
                binding_repo_id = _resolve_repo_id(
                    repo_id=binding.get("repo_id"),
                    workspace_path=binding.get("workspace_path"),
                    repo_id_by_workspace=repo_id_by_workspace,
                )
                if normalized_repo_id and binding_repo_id not in {
                    None,
                    normalized_repo_id,
                }:
                    continue
                channel_id = _normalize_optional_text(binding.get("channel_id"))
                if channel_id is None or channel_id in channels:
                    continue
                channels.append(channel_id)

            if not channels:
                return {"targets": 0, "published": 0}
            chunks = chunk_discord_message(
                format_discord_message(text),
                max_len=_DISCORD_MESSAGE_MAX_LEN,
                with_numbering=False,
            )
            if not chunks:
                chunks = [format_discord_message(text)]
            for channel_id in channels:
                targets += 1
                for index, chunk in enumerate(chunks, start=1):
                    digest = hashlib.sha256(
                        f"{correlation_id}:bound:{channel_id}:{index}".encode("utf-8")
                    ).hexdigest()[:24]
                    record_id = f"pma-notice:{digest}"
                    if await store.get_outbox(record_id) is not None:
                        continue
                    await store.enqueue_outbox(
                        DiscordOutboxRecord(
                            record_id=record_id,
                            channel_id=channel_id,
                            message_id=None,
                            operation="send",
                            payload_json={"content": chunk},
                            created_at=created_at,
                        )
                    )
                    published += 1
        finally:
            await store.close()
        return {"targets": targets, "published": published}

    async def _notify_telegram() -> dict[str, int]:
        from ..integrations.telegram.state import (
            OutboxRecord as TelegramOutboxRecord,
        )
        from ..integrations.telegram.state import TelegramStateStore, parse_topic_key

        targets = 0
        published = 0
        telegram_store = TelegramStateStore(
            _resolve_state_path(
                hub_root,
                raw_config,
                section="telegram_bot",
                default_state_file=TELEGRAM_STATE_FILE_DEFAULT,
            )
        )
        try:
            topics = await telegram_store.list_topics()
            for surface_key in sorted(topics):
                topic = topics[surface_key]
                if bool(getattr(topic, "pma_enabled", False)):
                    continue
                try:
                    chat_id, thread_id, scope = parse_topic_key(surface_key)
                except Exception:
                    continue
                base_key = f"{chat_id}:{thread_id or 'root'}"
                current_scope = await telegram_store.get_topic_scope(base_key)
                if scope != current_scope:
                    continue
                if not _binding_matches_workspace(
                    getattr(topic, "workspace_path", None), workspace_root
                ):
                    continue
                binding_repo_id = _resolve_repo_id(
                    repo_id=getattr(topic, "repo_id", None),
                    workspace_path=getattr(topic, "workspace_path", None),
                    repo_id_by_workspace=repo_id_by_workspace,
                )
                if normalized_repo_id and binding_repo_id not in {
                    None,
                    normalized_repo_id,
                }:
                    continue
                digest = hashlib.sha256(
                    f"{correlation_id}:bound:{chat_id}:{thread_id or 'root'}".encode(
                        "utf-8"
                    )
                ).hexdigest()[:24]
                record_id = f"pma-notice:{digest}"
                if await telegram_store.get_outbox(record_id) is not None:
                    continue
                await telegram_store.enqueue_outbox(
                    TelegramOutboxRecord(
                        record_id=record_id,
                        chat_id=chat_id,
                        thread_id=thread_id,
                        reply_to_message_id=None,
                        placeholder_message_id=None,
                        text=text,
                        created_at=created_at,
                        operation="send",
                        message_id=None,
                        outbox_key=f"pma-notice:{correlation_id}:{surface_key}:send",
                    )
                )
                targets += 1
                published += 1
        finally:
            await telegram_store.close()
        return {"targets": targets, "published": published}

    notify_by_source = {"discord": _notify_discord, "telegram": _notify_telegram}
    ordered_sources = [preferred_source] + [
        source for source in ("discord", "telegram") if source != preferred_source
    ]
    last_outcome = {"targets": 0, "published": 0}
    for source in ordered_sources:
        outcome = await notify_by_source[source]()
        if outcome.get("targets", 0) > 0:
            return outcome
        last_outcome = outcome
    return last_outcome


async def notify_primary_pma_chat_for_repo(
    *,
    hub_root: Path,
    repo_id: Optional[str],
    message: str,
    correlation_id: str,
) -> dict[str, int]:
    text = str(message or "").strip()
    normalized_repo_id = _normalize_optional_text(repo_id)
    if not text or normalized_repo_id is None:
        return {"targets": 0, "published": 0}

    try:
        raw_config = load_hub_config(hub_root).raw
    except Exception:
        raw_config = {}
    repo_id_by_workspace = _repo_id_by_workspace_path(hub_root)

    created_at = now_iso()
    discord_candidates: list[tuple[str, str]] = []
    from ..integrations.discord.rendering import (
        chunk_discord_message,
        format_discord_message,
    )
    from ..integrations.discord.state import DiscordStateStore
    from ..integrations.discord.state import OutboxRecord as DiscordOutboxRecord

    discord_store = DiscordStateStore(
        _resolve_state_path(
            hub_root,
            raw_config,
            section="discord_bot",
            default_state_file=DISCORD_STATE_FILE_DEFAULT,
        )
    )
    try:
        for binding in await discord_store.list_bindings():
            if not bool(binding.get("pma_enabled")):
                continue
            binding_repo_id = _resolve_repo_id(
                repo_id=binding.get("repo_id"),
                workspace_path=binding.get("workspace_path"),
                repo_id_by_workspace=repo_id_by_workspace,
            )
            prev_binding_repo_id = _resolve_repo_id(
                repo_id=binding.get("pma_prev_repo_id"),
                workspace_path=binding.get("pma_prev_workspace_path"),
                repo_id_by_workspace=repo_id_by_workspace,
            )
            prev_repo_id = _normalize_optional_text(binding.get("pma_prev_repo_id"))
            if normalized_repo_id not in {
                binding_repo_id,
                prev_repo_id,
                prev_binding_repo_id,
            }:
                continue
            channel_id = _normalize_optional_text(binding.get("channel_id"))
            updated_at = _normalize_optional_text(binding.get("updated_at")) or ""
            if channel_id is not None:
                discord_candidates.append((updated_at, channel_id))

        if discord_candidates:
            discord_candidates.sort(key=lambda item: (item[0], item[1]), reverse=True)
            channel_id = discord_candidates[0][1]
            chunks = chunk_discord_message(
                format_discord_message(text),
                max_len=_DISCORD_MESSAGE_MAX_LEN,
                with_numbering=False,
            )
            if not chunks:
                chunks = [format_discord_message(text)]
            published = 0
            for index, chunk in enumerate(chunks, start=1):
                digest = hashlib.sha256(
                    f"{correlation_id}:pma:{channel_id}:{index}".encode("utf-8")
                ).hexdigest()[:24]
                record_id = f"pma-escalation:{digest}"
                if await discord_store.get_outbox(record_id) is not None:
                    continue
                await discord_store.enqueue_outbox(
                    DiscordOutboxRecord(
                        record_id=record_id,
                        channel_id=channel_id,
                        message_id=None,
                        operation="send",
                        payload_json={"content": chunk},
                        created_at=created_at,
                    )
                )
                published += 1
            return {"targets": 1, "published": published}
    finally:
        await discord_store.close()

    from ..integrations.telegram.state import (
        OutboxRecord as TelegramOutboxRecord,
    )
    from ..integrations.telegram.state import TelegramStateStore, parse_topic_key

    telegram_store = TelegramStateStore(
        _resolve_state_path(
            hub_root,
            raw_config,
            section="telegram_bot",
            default_state_file=TELEGRAM_STATE_FILE_DEFAULT,
        )
    )
    try:
        topics = await telegram_store.list_topics()
        for surface_key in sorted(topics):
            topic = topics[surface_key]
            if not bool(getattr(topic, "pma_enabled", False)):
                continue
            try:
                chat_id, thread_id, scope = parse_topic_key(surface_key)
            except Exception:
                continue
            base_key = f"{chat_id}:{thread_id or 'root'}"
            current_scope = await telegram_store.get_topic_scope(base_key)
            if scope != current_scope:
                continue
            binding_repo_id = _resolve_repo_id(
                repo_id=getattr(topic, "repo_id", None),
                workspace_path=getattr(topic, "workspace_path", None),
                repo_id_by_workspace=repo_id_by_workspace,
            )
            prev_binding_repo_id = _resolve_repo_id(
                repo_id=getattr(topic, "pma_prev_repo_id", None),
                workspace_path=getattr(topic, "pma_prev_workspace_path", None),
                repo_id_by_workspace=repo_id_by_workspace,
            )
            prev_repo_id = _normalize_optional_text(
                getattr(topic, "pma_prev_repo_id", None)
            )
            if normalized_repo_id not in {
                binding_repo_id,
                prev_repo_id,
                prev_binding_repo_id,
            }:
                continue
            digest = hashlib.sha256(
                f"{correlation_id}:pma:{chat_id}:{thread_id or 'root'}".encode("utf-8")
            ).hexdigest()[:24]
            record_id = f"pma-escalation:{digest}"
            if await telegram_store.get_outbox(record_id) is not None:
                return {"targets": 1, "published": 0}
            await telegram_store.enqueue_outbox(
                TelegramOutboxRecord(
                    record_id=record_id,
                    chat_id=chat_id,
                    thread_id=thread_id,
                    reply_to_message_id=None,
                    placeholder_message_id=None,
                    text=text,
                    created_at=created_at,
                    operation="send",
                    message_id=None,
                    outbox_key=f"pma-escalation:{correlation_id}:{surface_key}:send",
                )
            )
            return {"targets": 1, "published": 1}
    finally:
        await telegram_store.close()
    return {"targets": 0, "published": 0}


__all__ = [
    "notify_preferred_bound_chat_for_workspace",
    "notify_primary_pma_chat_for_repo",
]



from __future__ import annotations

import asyncio
import logging
import uuid
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Optional

from fastapi import APIRouter, HTTPException, Request
from fastapi.responses import StreamingResponse

from .....agents.codex.harness import CodexHarness
from .....agents.opencode.harness import OpenCodeHarness
from .....core.orchestration import (
    SurfaceThreadMessageRequest,
    build_surface_orchestration_ingress,
)
from .....core.orchestration.runtime_thread_events import (
    RuntimeThreadRunEventState,
    normalize_runtime_thread_message,
    normalize_runtime_thread_raw_event,
)
from .....core.orchestration.turn_timeline import (
    iso_from_epoch_millis,
    persist_turn_timeline,
)
from .....core.pma_audit import PmaActionType
from .....core.pma_context import (
    PMA_MAX_TEXT,
    build_hub_snapshot,
    format_pma_prompt,
    load_pma_prompt,
)
from .....core.pma_lifecycle import PmaLifecycleRouter
from .....core.pma_queue import QueueItemState
from .....core.pma_state import PmaStateStore
from .....core.pma_transcripts import PmaTranscriptStore
from .....core.ports.run_event import (
    RUN_EVENT_DELTA_TYPE_ASSISTANT_MESSAGE,
    RUN_EVENT_DELTA_TYPE_ASSISTANT_STREAM,
    Completed,
    OutputDelta,
    RunEvent,
)
from .....core.time_utils import now_iso
from .....integrations.app_server.threads import PMA_KEY, PMA_OPENCODE_KEY
from .....integrations.github.context_injection import maybe_inject_github_context
from ...services.pma.common import (
    build_idempotency_key as service_build_idempotency_key,
)
from ...services.pma.common import pma_config_from_raw
from ..agents import _available_agents
from ..shared import SSE_HEADERS
from .automation_adapter import normalize_optional_text
from .publish import publish_automation_result
from .runtime_state import PmaRuntimeState
from .tail_stream import resolve_resume_after

logger = logging.getLogger(__name__)

PMA_TIMEOUT_SECONDS = 7200


def _normalize_optional_text(value: Any) -> Optional[str]:
    return normalize_optional_text(value)


def _get_pma_config(request: Request) -> dict[str, Any]:
    raw = getattr(request.app.state.config, "raw", {})
    return pma_config_from_raw(raw)


def _build_idempotency_key(
    *,
    lane_id: str,
    agent: Optional[str],
    model: Optional[str],
    reasoning: Optional[str],
    client_turn_id: Optional[str],
    message: str,
) -> str:
    return service_build_idempotency_key(
        lane_id=lane_id,
        agent=agent,
        model=model,
        reasoning=reasoning,
        client_turn_id=client_turn_id,
        message=message,
    )


def _truncate_text(text: str, max_len: int) -> str:
    if len(text) <= max_len:
        return text
    return text[: max_len - 3] + "..."


def _format_last_result(
    result: dict[str, Any], current: dict[str, Any]
) -> dict[str, Any]:

    status = result.get("status") or "error"
    message = result.get("message")
    detail = result.get("detail")
    text = message if isinstance(message, str) and message else detail
    summary = _truncate_text(text or "", PMA_MAX_TEXT)
    payload = {
        "status": status,
        "message": summary,
        "detail": (
            _truncate_text(detail or "", PMA_MAX_TEXT)
            if isinstance(detail, str)
            else None
        ),
        "client_turn_id": result.get("client_turn_id") or "",
        "agent": current.get("agent"),
        "thread_id": result.get("thread_id") or current.get("thread_id"),
        "turn_id": result.get("turn_id") or current.get("turn_id"),
        "started_at": current.get("started_at"),
        "finished_at": now_iso(),
    }
    delivery_status = _normalize_optional_text(result.get("delivery_status"))
    if delivery_status:
        payload["delivery_status"] = delivery_status
    delivery_outcome = result.get("delivery_outcome")
    if isinstance(delivery_outcome, dict):
        payload["delivery_outcome"] = dict(delivery_outcome)
    return payload


def _resolve_transcript_turn_id(result: dict[str, Any], current: dict[str, Any]) -> str:
    for candidate in (
        result.get("turn_id"),
        current.get("turn_id"),
        current.get("client_turn_id"),
    ):
        if isinstance(candidate, str) and candidate.strip():
            return candidate.strip()
    return f"local-{uuid.uuid4()}"


def _resolve_transcript_text(result: dict[str, Any]) -> str:
    message = result.get("message")
    if isinstance(message, str) and message.strip():
        return message
    detail = result.get("detail")
    if isinstance(detail, str) and detail.strip():
        return detail
    return ""


def _build_transcript_metadata(
    *,
    result: dict[str, Any],
    current: dict[str, Any],
    prompt_message: Optional[str],
    lifecycle_event: Optional[dict[str, Any]],
    model: Optional[str],
    reasoning: Optional[str],
    duration_ms: Optional[int],
    finished_at: str,
) -> dict[str, Any]:
    trigger = "lifecycle_event" if lifecycle_event else "user_prompt"
    metadata: dict[str, Any] = {
        "status": result.get("status") or "error",
        "agent": current.get("agent"),
        "thread_id": result.get("thread_id") or current.get("thread_id"),
        "turn_id": _resolve_transcript_turn_id(result, current),
        "client_turn_id": current.get("client_turn_id") or "",
        "lane_id": current.get("lane_id") or "",
        "trigger": trigger,
        "model": model,
        "reasoning": reasoning,
        "started_at": current.get("started_at"),
        "finished_at": finished_at,
        "duration_ms": duration_ms,
        "user_prompt": prompt_message or "",
    }
    if lifecycle_event:
        metadata["lifecycle_event"] = dict(lifecycle_event)
        metadata["event_id"] = lifecycle_event.get("event_id")
        metadata["event_type"] = lifecycle_event.get("event_type")
        metadata["repo_id"] = lifecycle_event.get("repo_id")
        metadata["run_id"] = lifecycle_event.get("run_id")
        metadata["event_timestamp"] = lifecycle_event.get("timestamp")
    return metadata


def _persist_timeline(
    *,
    hub_root: Path,
    execution_id: str,
    metadata: dict[str, Any],
    events: list[RunEvent],
) -> int:
    return persist_turn_timeline(
        hub_root,
        execution_id=execution_id,
        target_kind="lane",
        target_id=str(metadata.get("lane_id") or "").strip() or "pma:default",
        repo_id=_normalize_optional_text(metadata.get("repo_id")),
        run_id=_normalize_optional_text(metadata.get("run_id")),
        metadata=metadata,
        events=events,
    )


async def _persist_transcript(
    *,
    hub_root: Path,
    result: dict[str, Any],
    current: dict[str, Any],
    prompt_message: Optional[str],
    lifecycle_event: Optional[dict[str, Any]],
    model: Optional[str],
    reasoning: Optional[str],
    duration_ms: Optional[int],
    finished_at: str,
    timeline_events: Optional[list[RunEvent]] = None,
) -> Optional[dict[str, Any]]:

    store = PmaTranscriptStore(hub_root)
    assistant_text = _resolve_transcript_text(result)
    metadata = _build_transcript_metadata(
        result=result,
        current=current,
        prompt_message=prompt_message,
        lifecycle_event=lifecycle_event,
        model=model,
        reasoning=reasoning,
        duration_ms=duration_ms,
        finished_at=finished_at,
    )
    try:
        timeline_event_count = 0
        if timeline_events:
            timeline_event_count = _persist_timeline(
                hub_root=hub_root,
                execution_id=metadata["turn_id"],
                metadata=metadata,
                events=timeline_events,
            )
            if timeline_event_count:
                metadata["timeline_event_count"] = timeline_event_count
        pointer = store.write_transcript(
            turn_id=metadata["turn_id"],
            metadata=metadata,
            assistant_text=assistant_text,
        )
    except Exception:
        logger.exception("Failed to write PMA transcript")
        return None
    return {
        "turn_id": pointer.turn_id,
        "metadata_path": pointer.metadata_path,
        "content_path": pointer.content_path,
        "created_at": pointer.created_at,
    }


async def _finalize_result(
    runtime: PmaRuntimeState,
    result: dict[str, Any],
    *,
    request: Request,
    store: Optional[PmaStateStore] = None,
    prompt_message: Optional[str] = None,
    lifecycle_event: Optional[dict[str, Any]] = None,
    model: Optional[str] = None,
    reasoning: Optional[str] = None,
    timeline_events: Optional[list[RunEvent]] = None,
) -> None:

    async with await runtime.get_pma_lock():
        current_snapshot = dict(runtime.pma_current or {})
        runtime.pma_last_result = _format_last_result(result or {}, current_snapshot)
        runtime.pma_current = None
        runtime.pma_active = False
        runtime.pma_event = None
        runtime.pma_event_loop = None

    status = result.get("status") or "error"
    started_at = current_snapshot.get("started_at")
    duration_ms = None
    finished_at = now_iso()
    if started_at:
        try:
            start_dt = datetime.fromisoformat(started_at.replace("Z", "+00:00"))
            duration_ms = int(
                (datetime.now(timezone.utc) - start_dt).total_seconds() * 1000
            )
        except Exception:
            pass

    hub_root = request.app.state.config.root
    transcript_pointer = await _persist_transcript(
        hub_root=hub_root,
        result=result,
        current=current_snapshot,
        prompt_message=prompt_message,
        lifecycle_event=lifecycle_event,
        model=model,
        reasoning=reasoning,
        duration_ms=duration_ms,
        finished_at=finished_at,
        timeline_events=timeline_events,
    )
    if transcript_pointer is not None:
        runtime.pma_last_result = dict(runtime.pma_last_result or {})
        runtime.pma_last_result["transcript"] = transcript_pointer
        if not runtime.pma_last_result.get("turn_id"):
            runtime.pma_last_result["turn_id"] = transcript_pointer.get("turn_id")

    from .....core.logging_utils import log_event

    log_event(
        logger,
        logging.INFO,
        "pma.turn.completed",
        status=status,
        duration_ms=duration_ms,
        agent=current_snapshot.get("agent"),
        client_turn_id=current_snapshot.get("client_turn_id"),
        thread_id=runtime.pma_last_result.get("thread_id"),
        turn_id=runtime.pma_last_result.get("turn_id"),
        error=result.get("detail") if status == "error" else None,
    )

    safety_checker = runtime.get_safety_checker(hub_root, request)
    if status == "ok":
        action_type = PmaActionType.CHAT_COMPLETED
    elif status == "interrupted":
        action_type = PmaActionType.CHAT_INTERRUPTED
    else:
        action_type = PmaActionType.CHAT_FAILED

    safety_checker.record_action(
        action_type=action_type,
        agent=current_snapshot.get("agent"),
        thread_id=runtime.pma_last_result.get("thread_id"),
        turn_id=runtime.pma_last_result.get("turn_id"),
        client_turn_id=current_snapshot.get("client_turn_id"),
        details={"status": status, "duration_ms": duration_ms},
        status=status,
        error=result.get("detail") if status == "error" else None,
    )

    safety_checker.record_chat_result(
        agent=current_snapshot.get("agent") or "",
        status=status,
        error=result.get("detail") if status == "error" else None,
    )
    if lifecycle_event:
        safety_checker.record_reactive_result(
            status=status,
            error=result.get("detail") if status == "error" else None,
        )

    if store is not None:
        await runtime._persist_state(store)


async def _interrupt_active(
    runtime: PmaRuntimeState,
    request: Request,
    *,
    reason: str,
    source: str = "unknown",
) -> dict[str, Any]:
    event = await runtime.get_interrupt_event()
    event.set()
    current = await runtime.get_current_snapshot()
    agent_id = (current.get("agent") or "").strip().lower()
    thread_id = current.get("thread_id")
    turn_id = current.get("turn_id")
    client_turn_id = current.get("client_turn_id")
    hub_root = request.app.state.config.root

    from .....core.logging_utils import log_event

    log_event(
        logger,
        logging.INFO,
        "pma.turn.interrupted",
        agent=agent_id or None,
        client_turn_id=client_turn_id or None,
        thread_id=thread_id,
        turn_id=turn_id,
        reason=reason,
        source=source,
    )

    if agent_id == "opencode":
        supervisor = getattr(request.app.state, "opencode_supervisor", None)
        if supervisor is not None and thread_id:
            opencode_harness = OpenCodeHarness(supervisor)
            await opencode_harness.interrupt(hub_root, thread_id, turn_id)
    else:
        supervisor = getattr(request.app.state, "app_server_supervisor", None)
        events = getattr(request.app.state, "app_server_events", None)
        if supervisor is not None and events is not None and thread_id and turn_id:
            codex_harness = CodexHarness(supervisor, events)
            try:
                await codex_harness.interrupt(hub_root, thread_id, turn_id)
            except Exception:
                logger.exception("Failed to interrupt Codex turn")
    return {
        "status": "ok",
        "interrupted": bool(event.is_set()),
        "detail": reason,
        "agent": agent_id or None,
        "thread_id": thread_id,
        "turn_id": turn_id,
    }


def _cancel_background_task(task: asyncio.Task[Any], *, name: str) -> None:
    def _on_done(done_task: asyncio.Future[Any]) -> None:
        if isinstance(done_task, asyncio.Task):
            try:
                done_task.result()
            except asyncio.CancelledError:
                return
            except Exception:
                logger.exception("PMA task failed: %s", name)

    if task.done():
        _on_done(task)
        return

    task.add_done_callback(_on_done)
    task.cancel()


def _timeline_has_assistant_output(events: list[RunEvent]) -> bool:
    return any(
        isinstance(event, OutputDelta)
        and event.delta_type
        in {
            RUN_EVENT_DELTA_TYPE_ASSISTANT_MESSAGE,
            RUN_EVENT_DELTA_TYPE_ASSISTANT_STREAM,
        }
        for event in events
    )


async def _execute_app_server(
    supervisor: Any,
    events: Any,
    hub_root: Path,
    prompt: str,
    interrupt_event: asyncio.Event,
    *,
    model: Optional[str] = None,
    reasoning: Optional[str] = None,
    backend_thread_id: Optional[str] = None,
    thread_registry: Optional[Any] = None,
    thread_key: Optional[str] = None,
    on_meta: Optional[Any] = None,
) -> dict[str, Any]:
    client = await supervisor.get_client(hub_root)

    if backend_thread_id:
        thread_id = backend_thread_id
    elif thread_registry is not None and thread_key:
        thread_id = thread_registry.get_thread_id(thread_key)
    else:
        thread_id = None
    if thread_id:
        try:
            await client.thread_resume(thread_id)
        except Exception:
            thread_id = None

    if not thread_id:
        thread = await client.thread_start(str(hub_root))
        thread_id = thread.get("id")
        if not isinstance(thread_id, str) or not thread_id:
            raise HTTPException(
                status_code=502, detail="App-server did not return a thread id"
            )
        if thread_registry is not None and thread_key:
            thread_registry.set_thread_id(thread_key, thread_id)

    turn_kwargs: dict[str, Any] = {}
    if model:
        turn_kwargs["model"] = model
    if reasoning:
        turn_kwargs["effort"] = reasoning

    handle = await client.turn_start(
        thread_id,
        prompt,
        approval_policy="on-request",
        sandbox_policy="dangerFullAccess",
        **turn_kwargs,
    )
    register_turn = getattr(events, "register_turn", None)
    if callable(register_turn):
        try:
            await register_turn(thread_id, handle.turn_id)
        except Exception:
            logger.debug("pma chat register_turn failed", exc_info=True)
    codex_harness = CodexHarness(supervisor, events)
    if on_meta is not None:
        try:
            maybe = on_meta(thread_id, handle.turn_id)
            if asyncio.iscoroutine(maybe):
                await maybe
        except Exception:
            logger.exception("pma meta callback failed")

    if interrupt_event.is_set():
        try:
            await codex_harness.interrupt(hub_root, thread_id, handle.turn_id)
        except Exception:
            logger.exception("Failed to interrupt Codex turn")
        return {"status": "interrupted", "detail": "PMA chat interrupted"}

    turn_task = asyncio.create_task(handle.wait(timeout=None))
    timeout_task = asyncio.create_task(asyncio.sleep(PMA_TIMEOUT_SECONDS))
    interrupt_task = asyncio.create_task(interrupt_event.wait())
    try:
        done, _ = await asyncio.wait(
            {turn_task, timeout_task, interrupt_task},
            return_when=asyncio.FIRST_COMPLETED,
        )
        if timeout_task in done:
            try:
                await codex_harness.interrupt(hub_root, thread_id, handle.turn_id)
            except Exception:
                logger.exception("Failed to interrupt Codex turn")
            _cancel_background_task(turn_task, name="pma.app_server.turn.wait")
            return {"status": "error", "detail": "PMA chat timed out"}
        if interrupt_task in done:
            try:
                await codex_harness.interrupt(hub_root, thread_id, handle.turn_id)
            except Exception:
                logger.exception("Failed to interrupt Codex turn")
            _cancel_background_task(turn_task, name="pma.app_server.turn.wait")
            return {"status": "interrupted", "detail": "PMA chat interrupted"}
        turn_result = await turn_task
    finally:
        _cancel_background_task(timeout_task, name="pma.app_server.timeout.wait")
        _cancel_background_task(interrupt_task, name="pma.app_server.interrupt.wait")

    if getattr(turn_result, "errors", None):
        errors = turn_result.errors
        raise HTTPException(status_code=502, detail=errors[-1] if errors else "")

    output = str(getattr(turn_result, "final_message", "") or "").strip()
    if not output:
        output = "\n".join(getattr(turn_result, "agent_messages", []) or []).strip()
    raw_events = getattr(turn_result, "raw_events", []) or []
    timeline_events: list[RunEvent] = []
    list_events = getattr(events, "list_events", None)
    if callable(list_events):
        try:
            buffered = await list_events(thread_id, handle.turn_id)
        except Exception:
            buffered = []
        state = RuntimeThreadRunEventState()
        for entry in buffered:
            if not isinstance(entry, dict):
                continue
            message = entry.get("message")
            if not isinstance(message, dict):
                continue
            params = message.get("params")
            timeline_events.extend(
                normalize_runtime_thread_message(
                    str(message.get("method") or ""),
                    params if isinstance(params, dict) else {},
                    state,
                    timestamp=iso_from_epoch_millis(entry.get("received_at")),
                )
            )
    if not timeline_events and raw_events:
        state = RuntimeThreadRunEventState()
        for raw_event in raw_events:
            timeline_events.extend(
                await normalize_runtime_thread_raw_event(raw_event, state)
            )
    timeline_events.append(Completed(timestamp=now_iso(), final_message=output))
    return {
        "status": "ok",
        "message": output,
        "thread_id": thread_id,
        "backend_thread_id": thread_id,
        "turn_id": handle.turn_id,
        "raw_events": raw_events,
        "timeline_events": timeline_events,
    }


async def _execute_opencode(
    supervisor: Any,
    hub_root: Path,
    prompt: str,
    interrupt_event: asyncio.Event,
    *,
    model: Optional[str] = None,
    reasoning: Optional[str] = None,
    backend_session_id: Optional[str] = None,
    thread_registry: Optional[Any] = None,
    thread_key: Optional[str] = None,
    stall_timeout_seconds: Optional[float] = None,
    on_meta: Optional[Any] = None,
    part_handler: Optional[Any] = None,
) -> dict[str, Any]:
    from .....agents.opencode.runtime import (
        PERMISSION_ALLOW,
        build_turn_id,
        collect_opencode_output,
        extract_session_id,
        parse_message_response,
        split_model_id,
    )

    client = await supervisor.get_client(hub_root)
    session_id = backend_session_id
    if session_id is None and thread_registry is not None and thread_key:
        session_id = thread_registry.get_thread_id(thread_key)
    if not session_id:
        session = await client.create_session(directory=str(hub_root))
        session_id = extract_session_id(session, allow_fallback_id=True)
        if not isinstance(session_id, str) or not session_id:
            raise HTTPException(
                status_code=502, detail="OpenCode did not return a session id"
            )
        if thread_registry is not None and thread_key:
            thread_registry.set_thread_id(thread_key, session_id)
    if on_meta is not None:
        try:
            maybe = on_meta(session_id, build_turn_id(session_id))
            if asyncio.iscoroutine(maybe):
                await maybe
        except Exception:
            logger.exception("pma meta callback failed")

    opencode_harness = OpenCodeHarness(supervisor)
    if interrupt_event.is_set():
        await opencode_harness.interrupt(hub_root, session_id, None)
        return {"status": "interrupted", "detail": "PMA chat interrupted"}

    model_payload = split_model_id(model)
    await supervisor.mark_turn_started(hub_root)

    ready_event = asyncio.Event()
    timeline_state = RuntimeThreadRunEventState()
    timeline_events: list[RunEvent] = []

    async def _timeline_part_handler(
        part_type: str,
        part: dict[str, Any],
        delta_text: Optional[str],
    ) -> None:
        params: dict[str, Any] = {"part": dict(part or {})}
        if delta_text:
            params["delta"] = {"text": delta_text}
        timeline_events.extend(
            normalize_runtime_thread_message(
                "message.part.updated",
                params,
                timeline_state,
                timestamp=now_iso(),
            )
        )
        if part_handler is not None:
            maybe = part_handler(part_type, part, delta_text)
            if asyncio.iscoroutine(maybe):
                await maybe

    output_task = asyncio.create_task(
        collect_opencode_output(
            client,
            session_id=session_id,
            workspace_path=str(hub_root),
            model_payload=model_payload,
            permission_policy=PERMISSION_ALLOW,
            question_policy="auto_first_option",
            should_stop=interrupt_event.is_set,
            ready_event=ready_event,
            part_handler=_timeline_part_handler,
            stall_timeout_seconds=stall_timeout_seconds,
        )
    )
    try:
        await asyncio.wait_for(ready_event.wait(), timeout=2.0)
    except asyncio.TimeoutError:
        pass

    prompt_task = asyncio.create_task(
        client.prompt_async(
            session_id,
            message=prompt,
            model=model_payload,
            variant=reasoning,
        )
    )
    timeout_task = asyncio.create_task(asyncio.sleep(PMA_TIMEOUT_SECONDS))
    interrupt_task = asyncio.create_task(interrupt_event.wait())
    try:
        prompt_response = None
        try:
            prompt_response = await prompt_task
        except Exception as exc:
            interrupt_event.set()
            _cancel_background_task(output_task, name="pma.opencode.output.collect")
            await opencode_harness.interrupt(hub_root, session_id, None)
            raise HTTPException(status_code=502, detail=str(exc)) from exc

        done, _ = await asyncio.wait(
            {output_task, timeout_task, interrupt_task},
            return_when=asyncio.FIRST_COMPLETED,
        )
        if timeout_task in done:
            _cancel_background_task(output_task, name="pma.opencode.output.collect")
            await opencode_harness.interrupt(hub_root, session_id, None)
            return {"status": "error", "detail": "PMA chat timed out"}
        if interrupt_task in done:
            _cancel_background_task(output_task, name="pma.opencode.output.collect")
            await opencode_harness.interrupt(hub_root, session_id, None)
            return {"status": "interrupted", "detail": "PMA chat interrupted"}
        output_result = await output_task
        if (not output_result.text) and prompt_response is not None:
            fallback = parse_message_response(prompt_response)
            if fallback.text:
                output_result = type(output_result)(
                    text=fallback.text, error=fallback.error
                )
    finally:
        _cancel_background_task(timeout_task, name="pma.opencode.timeout.wait")
        _cancel_background_task(interrupt_task, name="pma.opencode.interrupt.wait")
        await supervisor.mark_turn_finished(hub_root)

    if output_result.error:
        raise HTTPException(status_code=502, detail=output_result.error)
    if (
        output_result.text
        and prompt_response is not None
        and not _timeline_has_assistant_output(timeline_events)
    ):
        completion_payload = (
            dict(prompt_response) if isinstance(prompt_response, dict) else {}
        )
        info = completion_payload.get("info")
        if not isinstance(info, dict):
            info = {}
        info = dict(info)
        if not isinstance(info.get("role"), str) or not str(info.get("role")).strip():
            info["role"] = "assistant"
        completion_payload["info"] = info
        if not isinstance(completion_payload.get("parts"), list):
            completion_payload["parts"] = [{"type": "text", "text": output_result.text}]
        timeline_events.extend(
            normalize_runtime_thread_message(
                "message.completed",
                completion_payload,
                timeline_state,
                timestamp=now_iso(),
            )
        )
    timeline_events.append(
        Completed(timestamp=now_iso(), final_message=output_result.text)
    )
    return {
        "status": "ok",
        "message": output_result.text,
        "thread_id": session_id,
        "backend_thread_id": session_id,
        "turn_id": build_turn_id(session_id),
        "timeline_events": timeline_events,
    }


async def _execute_queue_item(
    runtime: PmaRuntimeState,
    item: Any,
    request: Request,
) -> dict[str, Any]:
    hub_root = request.app.state.config.root
    payload = item.payload

    client_turn_id = payload.get("client_turn_id")
    message = payload.get("message", "")
    agent = payload.get("agent")
    model = _normalize_optional_text(payload.get("model"))
    reasoning = _normalize_optional_text(payload.get("reasoning"))
    lifecycle_event = payload.get("lifecycle_event")
    if not isinstance(lifecycle_event, dict):
        lifecycle_event = None
    wake_up = payload.get("wake_up")
    if not isinstance(wake_up, dict):
        wake_up = None
    automation_trigger = lifecycle_event is not None or wake_up is not None

    store = runtime.get_state_store(hub_root)
    defaults = _get_pma_config(request)
    started = False

    async def _finalize_queue_result_payload(
        result_payload: dict[str, Any],
        *,
        persist: bool = True,
    ) -> dict[str, Any]:
        payload_result = dict(result_payload or {})
        timeline_events = (
            payload_result.get("timeline_events")
            if isinstance(payload_result.get("timeline_events"), list)
            else None
        )
        payload_result.pop("timeline_events", None)
        if automation_trigger:
            try:
                payload_result.update(
                    await publish_automation_result(
                        request=request,
                        result=payload_result,
                        client_turn_id=(
                            _normalize_optional_text(client_turn_id) or None
                        ),
                        lifecycle_event=lifecycle_event,
                        wake_up=wake_up,
                    )
                )
            except Exception as exc:
                logger.exception(
                    "Failed publishing PMA automation result: client_turn_id=%s",
                    client_turn_id,
                )
                payload_result["delivery_status"] = "failed"
                payload_result["delivery_outcome"] = {
                    "published": 0,
                    "duplicates": 0,
                    "failed": 1,
                    "targets": 0,
                    "repo_id": None,
                    "correlation_id": (
                        _normalize_optional_text(client_turn_id)
                        or f"pma-{uuid.uuid4().hex[:12]}"
                    ),
                    "errors": [str(exc)],
                }
        if persist and started:
            await _finalize_result(
                runtime,
                payload_result,
                request=request,
                store=store,
                prompt_message=message,
                lifecycle_event=lifecycle_event,
                model=model,
                reasoning=reasoning,
                timeline_events=timeline_events,
            )
        return payload_result

    def _resolve_default_agent(available_ids: set[str], available_default: str) -> str:
        configured_default = defaults.get("default_agent")
        from .....agents.registry import validate_agent_id

        try:
            candidate = validate_agent_id(configured_default or "")
        except ValueError:
            candidate = None
        if candidate and candidate in available_ids:
            return candidate
        return available_default

    from .....agents.registry import validate_agent_id

    agents, available_default = _available_agents(request)
    available_ids = {entry.get("id") for entry in agents if isinstance(entry, dict)}

    try:
        agent_id = validate_agent_id(agent or "")
    except ValueError:
        agent_id = _resolve_default_agent(available_ids, available_default)

    safety_checker = runtime.get_safety_checker(hub_root, request)
    safety_check = safety_checker.check_chat_start(agent_id, message, client_turn_id)
    if not safety_check.allowed:
        detail = safety_check.reason or "PMA action blocked by safety check"
        if safety_check.details:
            detail = f"{detail}: {safety_check.details}"
        return await _finalize_queue_result_payload(
            {"status": "error", "detail": detail},
            persist=False,
        )

    started = await runtime.begin_turn(
        client_turn_id, store=store, lane_id=getattr(item, "lane_id", None)
    )
    if not started:
        detail = "Another PMA turn is already active; queue item was not started"
        logger.warning("PMA queue item rejected: %s", detail)
        return await _finalize_queue_result_payload(
            {
                "status": "error",
                "detail": detail,
                "client_turn_id": client_turn_id or "",
            },
            persist=False,
        )

    if not model and defaults.get("model"):
        model = defaults["model"]
    if not reasoning and defaults.get("reasoning"):
        reasoning = defaults["reasoning"]

    try:
        prompt_base = load_pma_prompt(hub_root)
        supervisor = getattr(request.app.state, "hub_supervisor", None)
        from .. import pma as pma_routes

        snapshot_builder = getattr(pma_routes, "build_hub_snapshot", build_hub_snapshot)
        github_context_injector = getattr(
            pma_routes,
            "maybe_inject_github_context",
            maybe_inject_github_context,
        )

        snapshot = await snapshot_builder(supervisor, hub_root=hub_root)
        prompt_state_key = PMA_OPENCODE_KEY if agent_id == "opencode" else PMA_KEY
        prompt = format_pma_prompt(
            prompt_base,
            snapshot,
            message,
            hub_root=hub_root,
            prompt_state_key=prompt_state_key,
        )
        prompt, _ = await github_context_injector(
            prompt_text=prompt,
            link_source_text=message,
            workspace_root=hub_root,
            logger=logger,
            event_prefix="web.pma.github_context",
            allow_cross_repo=True,
        )
    except Exception as exc:
        error_result = {
            "status": "error",
            "detail": str(exc),
            "client_turn_id": client_turn_id or "",
        }
        return await _finalize_queue_result_payload(error_result)

    interrupt_event = await runtime.get_interrupt_event()
    if interrupt_event.is_set():
        result = {"status": "interrupted", "detail": "PMA chat interrupted"}
        return await _finalize_queue_result_payload(result)

    async def _meta(thread_id: str, turn_id: str) -> None:
        await runtime.update_current(
            store=store,
            client_turn_id=client_turn_id or "",
            status="running",
            agent=agent_id,
            thread_id=thread_id,
            turn_id=turn_id,
        )

        safety_checker.record_action(
            action_type=PmaActionType.CHAT_STARTED,
            agent=agent_id,
            thread_id=thread_id,
            turn_id=turn_id,
            client_turn_id=client_turn_id,
            details={"message": message[:200]},
        )

        from .....core.logging_utils import log_event

        log_event(
            logger,
            logging.INFO,
            "pma.turn.started",
            agent=agent_id,
            client_turn_id=client_turn_id or None,
            thread_id=thread_id,
            turn_id=turn_id,
        )

    supervisor = getattr(request.app.state, "app_server_supervisor", None)
    events = getattr(request.app.state, "app_server_events", None)
    opencode = getattr(request.app.state, "opencode_supervisor", None)
    registry = getattr(request.app.state, "app_server_threads", None)
    stall_timeout_seconds = None
    try:
        stall_timeout_seconds = (
            request.app.state.config.opencode.session_stall_timeout_seconds
        )
    except Exception:
        stall_timeout_seconds = None

    ingress = build_surface_orchestration_ingress(
        event_sink=lambda orchestration_event: logger.info(
            "web.pma.%s surface=%s target_kind=%s target_id=%s status=%s meta=%s",
            orchestration_event.event_type,
            orchestration_event.surface_kind,
            orchestration_event.target_kind,
            orchestration_event.target_id,
            orchestration_event.status,
            orchestration_event.metadata,
        )
    )

    async def _resolve_no_flow(
        _request: SurfaceThreadMessageRequest,
    ) -> None:
        return None

    async def _submit_flow_reply(
        _request: SurfaceThreadMessageRequest, _flow_target: Any
    ) -> dict[str, Any]:
        raise RuntimeError("PMA web ingress does not route ticket_flow replies")

    async def _submit_thread_message(
        _request: SurfaceThreadMessageRequest,
    ) -> dict[str, Any]:
        if agent_id == "opencode":
            if opencode is None:
                return {"status": "error", "detail": "OpenCode unavailable"}
            return await _execute_opencode(
                opencode,
                hub_root,
                prompt,
                interrupt_event,
                model=model,
                reasoning=reasoning,
                thread_registry=registry,
                thread_key=PMA_OPENCODE_KEY,
                stall_timeout_seconds=stall_timeout_seconds,
                on_meta=_meta,
            )
        if supervisor is None or events is None:
            return {"status": "error", "detail": "App-server unavailable"}
        return await _execute_app_server(
            supervisor,
            events,
            hub_root,
            prompt,
            interrupt_event,
            model=model,
            reasoning=reasoning,
            thread_registry=registry,
            thread_key=PMA_KEY,
            on_meta=_meta,
        )

    try:
        ingress_result = await ingress.submit_message(
            SurfaceThreadMessageRequest(
                surface_kind="web",
                workspace_root=hub_root,
                prompt_text=message,
                agent_id=agent_id,
                pma_enabled=True,
                metadata={"client_turn_id": client_turn_id or ""},
            ),
            resolve_paused_flow_target=_resolve_no_flow,
            submit_flow_reply=_submit_flow_reply,
            submit_thread_message=_submit_thread_message,
        )
        result = dict(ingress_result.thread_result or {})
    except Exception as exc:
        error_result = {
            "status": "error",
            "detail": str(exc),
            "client_turn_id": client_turn_id or "",
        }
        await _finalize_queue_result_payload(error_result)
        raise

    result = dict(result or {})
    result["client_turn_id"] = client_turn_id or ""
    return await _finalize_queue_result_payload(result)


def _require_pma_enabled(request: Request) -> None:
    pma_config = _get_pma_config(request)
    if not pma_config.get("enabled", True):
        raise HTTPException(status_code=404, detail="PMA is disabled")


class _AppRequest:
    def __init__(self, app: Any) -> None:
        self.app = app


async def _ensure_lane_worker_for_app(
    runtime: PmaRuntimeState, app: Any, lane_id: str
) -> None:
    await runtime.ensure_lane_worker(
        lane_id,
        _AppRequest(app),
        lambda item: _execute_queue_item(runtime, item, _AppRequest(app)),
    )


async def _stop_lane_worker_for_app(
    runtime: PmaRuntimeState, app: Any, lane_id: str
) -> None:
    _ = app
    await runtime.stop_lane_worker(lane_id)


async def _stop_all_lane_workers_for_app(runtime: PmaRuntimeState, app: Any) -> None:
    _ = app
    await runtime.stop_all_lane_workers()


def build_chat_runtime_router(
    router: APIRouter,
    get_runtime_state: Any,
) -> None:
    """Build PMA chat runtime routes.

    This includes:
    - /active - Get current PMA status
    - /chat - Submit a PMA chat message
    - /interrupt - Interrupt running PMA turn
    - /stop - Stop a PMA lane
    - /new - Create new PMA session
    - /reset - Reset PMA state
    - /compact - Compact PMA history
    - /thread/reset - Reset PMA thread
    - /queue - Get queue summary
    - /queue/{lane_id} - Get lane queue items
    - /turns/{turn_id}/events - Stream turn events
    """

    @router.get("/active")
    async def pma_active_status(
        request: Request, client_turn_id: Optional[str] = None
    ) -> dict[str, Any]:
        runtime = get_runtime_state()
        async with await runtime.get_pma_lock():
            current = dict(runtime.pma_current or {})
            last_result = dict(runtime.pma_last_result or {})
            active = bool(runtime.pma_active)
        store = runtime.get_state_store(request.app.state.config.root)
        disk_state = store.load(ensure_exists=True)
        if isinstance(disk_state, dict):
            disk_current = (
                disk_state.get("current")
                if isinstance(disk_state.get("current"), dict)
                else {}
            )
            disk_last = (
                disk_state.get("last_result")
                if isinstance(disk_state.get("last_result"), dict)
                else {}
            )
            if not current and disk_current:
                current = dict(disk_current)
            if not last_result and disk_last:
                last_result = dict(disk_last)
            if not active and disk_state.get("active"):
                active = True
        if client_turn_id:
            if last_result.get("client_turn_id") != client_turn_id:
                last_result = {}
            if current.get("client_turn_id") != client_turn_id:
                current = {}
        return {"active": active, "current": current, "last_result": last_result}

    @router.post("/chat")
    async def pma_chat(request: Request):
        pma_config = _get_pma_config(request)
        body = await request.json()
        message = (body.get("message") or "").strip()
        stream = bool(body.get("stream", False))
        agent = _normalize_optional_text(body.get("agent"))
        model = _normalize_optional_text(body.get("model"))
        reasoning = _normalize_optional_text(body.get("reasoning"))
        client_turn_id = (body.get("client_turn_id") or "").strip() or None

        if not message:
            raise HTTPException(status_code=400, detail="message is required")
        max_text_chars = int(pma_config.get("max_text_chars", 0) or 0)
        if max_text_chars > 0 and len(message) > max_text_chars:
            raise HTTPException(
                status_code=400,
                detail=(
                    f"message exceeds max_text_chars ({max_text_chars} characters)"
                ),
            )

        runtime = get_runtime_state()
        hub_root = request.app.state.config.root
        queue = runtime.get_pma_queue(hub_root)

        lane_id = "pma:default"
        idempotency_key = _build_idempotency_key(
            lane_id=lane_id,
            agent=agent,
            model=model,
            reasoning=reasoning,
            client_turn_id=client_turn_id,
            message=message,
        )

        payload = {
            "message": message,
            "agent": agent,
            "model": model,
            "reasoning": reasoning,
            "client_turn_id": client_turn_id,
            "stream": stream,
            "hub_root": str(hub_root),
        }

        item, dupe_reason = await queue.enqueue(lane_id, idempotency_key, payload)
        if dupe_reason:
            logger.info("Duplicate PMA turn: %s", dupe_reason)

        if item.state == QueueItemState.DEDUPED:
            return {
                "status": "ok",
                "message": "Duplicate request - already processing",
                "deduped": True,
            }

        result_future = asyncio.get_running_loop().create_future()
        runtime.item_futures[item.item_id] = result_future

        await runtime.ensure_lane_worker(
            lane_id,
            request,
            lambda item: _execute_queue_item(runtime, item, request),
        )

        try:
            result = await asyncio.wait_for(result_future, timeout=PMA_TIMEOUT_SECONDS)
        except asyncio.TimeoutError:
            return {"status": "error", "detail": "PMA chat timed out"}
        except Exception:
            logger.exception("PMA chat error")
            return {
                "status": "error",
                "detail": "An error occurred processing your request",
            }

        return result

    @router.post("/interrupt")
    async def pma_interrupt(request: Request) -> dict[str, Any]:
        runtime = get_runtime_state()
        return await _interrupt_active(
            runtime, request, reason="PMA chat interrupted", source="user_request"
        )

    @router.post("/stop")
    async def pma_stop(request: Request) -> dict[str, Any]:
        body = await request.json() if request.headers.get("content-type") else {}
        lane_id = (body.get("lane_id") or "pma:default").strip()
        hub_root = request.app.state.config.root
        lifecycle_router = PmaLifecycleRouter(hub_root)

        runtime = get_runtime_state()
        result = await lifecycle_router.stop(lane_id=lane_id)

        if result.status != "ok":
            raise HTTPException(status_code=500, detail=result.error)

        await runtime.stop_lane_worker(lane_id)

        await _interrupt_active(
            runtime, request, reason="Lane stopped", source="user_request"
        )

        return {
            "status": result.status,
            "message": result.message,
            "artifact_path": (
                str(result.artifact_path) if result.artifact_path else None
            ),
            "details": result.details,
        }

    @router.post("/new")
    async def new_pma_session(request: Request) -> dict[str, Any]:
        body = await request.json()
        agent = _normalize_optional_text(body.get("agent"))
        lane_id = (body.get("lane_id") or "pma:default").strip()

        hub_root = request.app.state.config.root
        lifecycle_router = PmaLifecycleRouter(hub_root)

        result = await lifecycle_router.new(agent=agent, lane_id=lane_id)

        if result.status != "ok":
            raise HTTPException(status_code=500, detail=result.error)

        return {
            "status": result.status,
            "message": result.message,
            "artifact_path": (
                str(result.artifact_path) if result.artifact_path else None
            ),
            "details": result.details,
        }

    @router.post("/reset")
    async def reset_pma_session(request: Request) -> dict[str, Any]:
        body = await request.json() if request.headers.get("content-type") else {}
        raw_agent = (body.get("agent") or "").strip().lower()
        agent = raw_agent or None

        hub_root = request.app.state.config.root
        lifecycle_router = PmaLifecycleRouter(hub_root)

        result = await lifecycle_router.reset(agent=agent)

        if result.status != "ok":
            raise HTTPException(status_code=500, detail=result.error)

        return {
            "status": result.status,
            "message": result.message,
            "artifact_path": (
                str(result.artifact_path) if result.artifact_path else None
            ),
            "details": result.details,
        }

    @router.post("/compact")
    async def compact_pma_history(request: Request) -> dict[str, Any]:
        body = await request.json()
        summary = (body.get("summary") or "").strip()
        agent = _normalize_optional_text(body.get("agent"))
        thread_id = _normalize_optional_text(body.get("thread_id"))

        if not summary:
            raise HTTPException(status_code=400, detail="summary is required")

        hub_root = request.app.state.config.root
        lifecycle_router = PmaLifecycleRouter(hub_root)

        result = await lifecycle_router.compact(
            summary=summary, agent=agent, thread_id=thread_id
        )

        if result.status != "ok":
            raise HTTPException(status_code=500, detail=result.error)

        return {
            "status": result.status,
            "message": result.message,
            "artifact_path": (
                str(result.artifact_path) if result.artifact_path else None
            ),
            "details": result.details,
        }

    @router.post("/thread/reset")
    async def reset_pma_thread(request: Request) -> dict[str, Any]:
        body = await request.json()
        raw_agent = (body.get("agent") or "").strip().lower()
        agent = raw_agent or None

        hub_root = request.app.state.config.root
        lifecycle_router = PmaLifecycleRouter(hub_root)

        result = await lifecycle_router.reset(agent=agent)

        if result.status != "ok":
            raise HTTPException(status_code=500, detail=result.error)

        return {
            "status": result.status,
            "cleared": result.details.get("cleared_threads", []),
            "artifact_path": (
                str(result.artifact_path) if result.artifact_path else None
            ),
        }

    @router.get("/queue")
    async def pma_queue_status(request: Request) -> dict[str, Any]:
        runtime = get_runtime_state()
        queue = runtime.get_pma_queue(request.app.state.config.root)
        summary = await queue.get_queue_summary()
        return summary

    @router.get("/queue/{lane_id:path}")
    async def pma_lane_queue_status(request: Request, lane_id: str) -> dict[str, Any]:
        runtime = get_runtime_state()
        queue = runtime.get_pma_queue(request.app.state.config.root)
        items = await queue.list_items(lane_id)
        return {
            "lane_id": lane_id,
            "items": [
                {
                    "item_id": item.item_id,
                    "state": item.state.value,
                    "enqueued_at": item.enqueued_at,
                    "started_at": item.started_at,
                    "finished_at": item.finished_at,
                    "error": item.error,
                    "dedupe_reason": item.dedupe_reason,
                }
                for item in items
            ],
        }

    @router.get("/turns/{turn_id}/events")
    async def stream_pma_turn_events(
        turn_id: str,
        request: Request,
        thread_id: str,
        agent: str = "codex",
        since_event_id: Optional[int] = None,
    ):
        agent_id = (agent or "").strip().lower()
        resume_after = resolve_resume_after(request, since_event_id)
        if agent_id == "codex":
            events = getattr(request.app.state, "app_server_events", None)
            if events is None:
                raise HTTPException(status_code=404, detail="Codex events unavailable")
            if not thread_id:
                raise HTTPException(status_code=400, detail="thread_id is required")
            return StreamingResponse(
                events.stream(thread_id, turn_id, after_id=(resume_after or 0)),
                media_type="text/event-stream",
                headers=SSE_HEADERS,
            )
        if agent_id == "opencode":
            if not thread_id:
                raise HTTPException(status_code=400, detail="thread_id is required")
            supervisor = getattr(request.app.state, "opencode_supervisor", None)
            if supervisor is None:
                raise HTTPException(status_code=404, detail="OpenCode unavailable")
            harness = OpenCodeHarness(supervisor)
            return StreamingResponse(
                harness.stream_events(
                    request.app.state.config.root, thread_id, turn_id
                ),
                media_type="text/event-stream",
                headers=SSE_HEADERS,
            )
        raise HTTPException(status_code=404, detail="Unknown agent")



import json
from pathlib import Path

from typer.testing import CliRunner

from codex_autorunner.bootstrap import seed_hub_files
from codex_autorunner.cli import app
from codex_autorunner.integrations.chat.channel_directory import ChannelDirectoryStore

runner = CliRunner()


def test_chat_channels_list_shows_entries(tmp_path: Path) -> None:
    seed_hub_files(tmp_path, force=True)
    store = ChannelDirectoryStore(tmp_path)
    store.record_seen(
        "telegram",
        "-1001",
        "77",
        "Team Room / Ops",
        {"chat_type": "supergroup"},
    )

    result = runner.invoke(app, ["chat", "channels", "list", "--path", str(tmp_path)])

    assert result.exit_code == 0
    output = result.output
    assert "Chat channel directory entries:" in output
    assert "telegram:-1001:77" in output
    assert "Team Room / Ops" in output


def test_chat_channels_list_query_filters_with_json_output(tmp_path: Path) -> None:
    seed_hub_files(tmp_path, force=True)
    store = ChannelDirectoryStore(tmp_path)
    store.record_seen(
        "discord",
        "channel-1",
        "guild-1",
        "CAR HQ / #general",
        {"guild_id": "guild-1"},
    )
    store.record_seen(
        "telegram",
        "-1002",
        "42",
        "Team Room / Build",
        {"chat_type": "supergroup"},
    )

    result = runner.invoke(
        app,
        [
            "chat",
            "channels",
            "list",
            "--query",
            "car hq",
            "--json",
            "--path",
            str(tmp_path),
        ],
    )

    assert result.exit_code == 0
    payload = json.loads(result.output)
    entries = payload["entries"]
    assert len(entries) == 1
    assert entries[0]["key"] == "discord:channel-1:guild-1"
    assert entries[0]["display"] == "CAR HQ / #general"



"""
Agent harness support routes (models + event streaming).
"""

from __future__ import annotations

from typing import Any, Optional

from fastapi import APIRouter, HTTPException, Request
from fastapi.responses import StreamingResponse

from ....agents.codex.harness import CodexHarness
from ....agents.opencode.harness import OpenCodeHarness
from ....agents.opencode.supervisor import OpenCodeSupervisorError
from ....agents.registry import get_available_agents
from ....agents.types import ModelCatalog
from ....core.orchestration.catalog import map_agent_capabilities
from ..services.validation import normalize_agent_id
from .shared import SSE_HEADERS


def _normalize_path_agent_id(agent: str) -> str:
    # Path segments that decode to blank/whitespace are malformed and should
    # not silently fall back to the default agent.
    if not isinstance(agent, str) or not agent.strip():
        raise HTTPException(status_code=404, detail="Unknown agent")
    return normalize_agent_id(agent)


def _available_agents(request: Request) -> tuple[list[dict[str, Any]], str]:
    agents: list[dict[str, Any]] = []
    default_agent: Optional[str] = None

    available = get_available_agents(request.app.state)

    for agent_id, descriptor in available.items():
        agent_data: dict[str, Any] = {
            "id": agent_id,
            "name": descriptor.name,
            "capabilities": sorted(map_agent_capabilities(descriptor.capabilities)),
        }
        if agent_id == "codex":
            agent_data["protocol_version"] = "2.0"
        if agent_id == "opencode":
            supervisor = getattr(request.app.state, "opencode_supervisor", None)
            if supervisor and hasattr(supervisor, "_handles"):
                handles = supervisor._handles
                if handles:
                    first_handle = next(iter(handles.values()), None)
                    if first_handle:
                        version = getattr(first_handle, "version", None)
                        if version:
                            agent_data["version"] = str(version)
        agents.append(agent_data)
        if default_agent is None:
            default_agent = agent_id

    if not agents:
        agents = [
            {
                "id": "codex",
                "name": "Codex",
                "protocol_version": "2.0",
                "capabilities": [],
            }
        ]
        default_agent = "codex"

    return agents, default_agent or "codex"


def _serialize_model_catalog(catalog: ModelCatalog) -> dict[str, Any]:
    return {
        "default_model": catalog.default_model,
        "models": [
            {
                "id": model.id,
                "display_name": model.display_name,
                "supports_reasoning": model.supports_reasoning,
                "reasoning_options": list(model.reasoning_options),
            }
            for model in catalog.models
        ],
    }


def build_agents_routes() -> APIRouter:
    router = APIRouter()

    @router.get("/api/agents")
    def list_agents(request: Request) -> dict[str, Any]:
        agents, default_agent = _available_agents(request)
        return {"agents": agents, "default": default_agent}

    @router.get("/api/agents/{agent}/models")
    async def list_agent_models(agent: str, request: Request):
        agent_id = _normalize_path_agent_id(agent)
        engine = request.app.state.engine
        if agent_id == "codex":
            supervisor = request.app.state.app_server_supervisor
            events = request.app.state.app_server_events
            if supervisor is None:
                raise HTTPException(status_code=404, detail="Codex harness unavailable")
            codex_harness = CodexHarness(supervisor, events)
            catalog = await codex_harness.model_catalog(engine.repo_root)
            return _serialize_model_catalog(catalog)
        if agent_id == "opencode":
            supervisor = getattr(request.app.state, "opencode_supervisor", None)
            if supervisor is None:
                raise HTTPException(
                    status_code=404, detail="OpenCode harness unavailable"
                )
            try:
                opencode_harness = OpenCodeHarness(supervisor)
                catalog = await opencode_harness.model_catalog(engine.repo_root)
                return _serialize_model_catalog(catalog)
            except OpenCodeSupervisorError as exc:
                raise HTTPException(status_code=502, detail=str(exc)) from exc
            except Exception as exc:
                raise HTTPException(status_code=502, detail=str(exc)) from exc
        raise HTTPException(status_code=404, detail="Unknown agent")

    @router.get("/api/agents/{agent}/turns/{turn_id}/events")
    async def stream_agent_turn_events(
        agent: str,
        turn_id: str,
        request: Request,
        thread_id: Optional[str] = None,
        since_event_id: Optional[int] = None,
    ):
        agent_id = _normalize_path_agent_id(agent)
        resume_after = since_event_id
        if resume_after is None:
            last_event_id = request.headers.get("Last-Event-ID")
            if last_event_id:
                try:
                    resume_after = int(last_event_id)
                except ValueError:
                    resume_after = None
        if agent_id == "codex":
            events = getattr(request.app.state, "app_server_events", None)
            if events is None:
                raise HTTPException(status_code=404, detail="Codex events unavailable")
            if not thread_id:
                raise HTTPException(status_code=400, detail="thread_id is required")
            return StreamingResponse(
                events.stream(thread_id, turn_id, after_id=(resume_after or 0)),
                media_type="text/event-stream",
                headers=SSE_HEADERS,
            )
        if agent_id == "opencode":
            if not thread_id:
                raise HTTPException(status_code=400, detail="thread_id is required")
            supervisor = getattr(request.app.state, "opencode_supervisor", None)
            if supervisor is None:
                raise HTTPException(
                    status_code=404, detail="OpenCode events unavailable"
                )
            harness = OpenCodeHarness(supervisor)
            return StreamingResponse(
                harness.stream_events(
                    request.app.state.engine.repo_root, thread_id, turn_id
                ),
                media_type="text/event-stream",
                headers=SSE_HEADERS,
            )
        raise HTTPException(status_code=404, detail="Unknown agent")

    return router


__all__ = ["build_agents_routes"]



from __future__ import annotations

import asyncio
import json
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Optional

from fastapi import APIRouter, HTTPException, Request
from fastapi.responses import StreamingResponse

from .....core.app_server_logging import AppServerEventFormatter
from .....core.pma_thread_store import PmaThreadStore
from .....core.redaction import redact_text
from ..shared import SSE_HEADERS
from .automation_adapter import normalize_optional_text
from .managed_threads import (
    _serialize_thread_target,
    build_managed_thread_orchestration_service,
)


def coerce_dict(value: Any) -> dict[str, Any]:
    return value if isinstance(value, dict) else {}


def parse_iso_datetime(value: Any) -> Optional[datetime]:
    if not isinstance(value, str) or not value.strip():
        return None
    try:
        return datetime.fromisoformat(value.replace("Z", "+00:00"))
    except ValueError:
        return None


def parse_tail_duration_seconds(value: Optional[str]) -> Optional[int]:
    if value is None:
        return None
    raw = value.strip().lower()
    if not raw:
        raise HTTPException(status_code=400, detail="since must not be empty")
    multipliers = {"s": 1, "m": 60, "h": 3600, "d": 86400, "w": 604800}
    total_seconds = 0
    idx = 0
    size = len(raw)
    while idx < size:
        start = idx
        while idx < size and raw[idx].isdigit():
            idx += 1
        if start == idx or idx >= size:
            raise HTTPException(
                status_code=400,
                detail=(
                    "Invalid since duration. Use forms like 30s, 5m, 2h, 1d, "
                    "or combined 1h30m."
                ),
            )
        amount_text = raw[start:idx]
        if len(amount_text) > 9:
            raise HTTPException(
                status_code=400, detail="since duration component is too large"
            )
        multiplier = multipliers.get(raw[idx])
        if multiplier is None:
            raise HTTPException(
                status_code=400,
                detail=(
                    "Invalid since duration. Use forms like 30s, 5m, 2h, 1d, "
                    "or combined 1h30m."
                ),
            )
        idx += 1
        total_seconds += int(amount_text) * multiplier
    if total_seconds <= 0:
        raise HTTPException(status_code=400, detail="since must be > 0")
    return total_seconds


def since_ms_from_duration(value: Optional[str]) -> Optional[int]:
    seconds = parse_tail_duration_seconds(value)
    if seconds is None:
        return None
    return int((datetime.now(timezone.utc).timestamp() - seconds) * 1000)


def normalize_tail_level(level: Optional[str]) -> str:
    normalized = (level or "info").strip().lower() or "info"
    if normalized not in {"info", "debug"}:
        raise HTTPException(status_code=400, detail="level must be info or debug")
    return normalized


def resolve_resume_after(
    request: Request, since_event_id: Optional[int]
) -> Optional[int]:
    if since_event_id is not None:
        if since_event_id < 0:
            raise HTTPException(status_code=400, detail="since_event_id must be >= 0")
        return since_event_id
    last_event_id = request.headers.get("Last-Event-ID")
    if not last_event_id:
        return None
    try:
        parsed = int(last_event_id)
    except ValueError as exc:
        raise HTTPException(
            status_code=400, detail="Invalid Last-Event-ID header"
        ) from exc
    if parsed < 0:
        raise HTTPException(status_code=400, detail="Last-Event-ID must be >= 0")
    return parsed


def iso_from_event_ms(value: Any) -> Optional[str]:
    if not isinstance(value, (int, float)) or value <= 0:
        return None
    return datetime.fromtimestamp(value / 1000.0, tz=timezone.utc).isoformat()


def truncate_text(text: str, max_length: int) -> str:
    if len(text) <= max_length:
        return text
    return text[: max_length - 3] + "..."


def _redact_nested(value: Any) -> Any:
    if isinstance(value, str):
        return redact_text(value)
    if isinstance(value, dict):
        return {str(k): _redact_nested(v) for k, v in value.items()}
    if isinstance(value, list):
        return [_redact_nested(item) for item in value]
    return value


def _tail_event_type_from_message(message: Any) -> str:
    payload = coerce_dict(message)
    method = str(payload.get("method") or "").strip().lower()
    params = coerce_dict(payload.get("params"))
    item = coerce_dict(params.get("item"))

    if method == "turn/completed":
        status = str(params.get("status") or "").strip().lower()
        if status in {"interrupt", "interrupted"}:
            return "turn_interrupted"
        if status in {"error", "failed"}:
            return "turn_failed"
        return "turn_completed"
    if method == "error":
        return "turn_failed"
    if method in {
        "item/commandexecution/requestapproval",
        "item/filechange/requestapproval",
    }:
        return "tool_started"
    if method == "item/toolcall/start":
        return "tool_started"
    if method == "item/toolcall/end":
        success = params.get("success")
        if success is False:
            return "tool_failed"
        return "tool_completed"
    if method == "item/completed":
        item_type = str(item.get("type") or "").strip().lower()
        if item_type in {"commandexecution", "filechange", "tool"}:
            exit_code = item.get("exitCode")
            if isinstance(exit_code, int) and exit_code != 0:
                return "tool_failed"
            return "tool_completed"
    if "reasoning" in method:
        return "assistant_update"
    return "progress"


def _should_suppress_tail_event(message: Any) -> bool:
    payload = coerce_dict(message)
    method = str(payload.get("method") or "").strip().lower()
    params = coerce_dict(payload.get("params"))
    item = coerce_dict(params.get("item"))

    if method in {
        "item/agentmessage/delta",
        "item/plan/delta",
        "turn/plan/updated",
    }:
        return True
    if method == "item/completed":
        item_type = str(item.get("type") or "").strip().lower()
        if item_type == "agentmessage":
            return True
    return False


_NO_STREAM_AVAILABLE_IDLE_SECONDS = 15
_LIKELY_HUNG_IDLE_SECONDS = 90


def _truncate_tool_name(value: Any) -> str | None:
    text = normalize_optional_text(value)
    if text is None:
        return None
    return truncate_text(text, 80)


def _tool_metadata_from_message(message: Any) -> tuple[str | None, str | None]:
    payload = coerce_dict(message)
    method = str(payload.get("method") or "").strip().lower()
    params = coerce_dict(payload.get("params"))
    item = coerce_dict(params.get("item"))
    tool_name = _truncate_tool_name(
        item.get("name")
        or item.get("toolName")
        or item.get("command")
        or params.get("tool_name")
        or params.get("toolName")
        or params.get("command")
        or params.get("name")
    )
    tool_state: str | None = None
    if method in {
        "item/commandexecution/requestapproval",
        "item/filechange/requestapproval",
        "item/toolcall/start",
    }:
        tool_state = "started"
    elif method == "item/toolcall/end":
        tool_state = "failed" if params.get("success") is False else "completed"
    elif method == "item/completed":
        item_type = str(item.get("type") or "").strip().lower()
        if item_type in {"commandexecution", "filechange", "tool"}:
            exit_code = item.get("exitCode")
            tool_state = (
                "failed"
                if isinstance(exit_code, int) and exit_code != 0
                else "completed"
            )
    return tool_name, tool_state


def _parse_inline_sse(raw_event: str) -> tuple[str, dict[str, Any]]:
    event_name = "message"
    data_lines: list[str] = []
    for raw_line in str(raw_event).splitlines():
        line = raw_line.rstrip("\n")
        if not line:
            continue
        if line.startswith("event:"):
            event_name = line.split(":", 1)[1].strip() or "message"
        elif line.startswith("data:"):
            data_lines.append(line.split(":", 1)[1].lstrip())
    payload: dict[str, Any] = {}
    data = "\n".join(data_lines)
    if data:
        try:
            parsed = json.loads(data)
        except json.JSONDecodeError:
            payload = {}
        else:
            payload = coerce_dict(parsed)
    return event_name, payload


def _zeroclaw_tail_event_from_raw(
    *,
    event_id: int,
    raw_event: str,
    published_at: str | None,
    since_ms: Optional[int],
) -> dict[str, Any] | None:
    event_name, payload = _parse_inline_sse(raw_event)
    message = coerce_dict(payload.get("message"))
    method = str(message.get("method") or "").strip().lower()
    params = coerce_dict(message.get("params"))
    text = redact_text(str(params.get("text") or "")).strip()
    if not method:
        return None
    received_at = normalize_optional_text(published_at)
    received_at_ms = None
    if received_at is not None:
        dt = parse_iso_datetime(received_at)
        if dt is not None:
            received_at_ms = int(dt.timestamp() * 1000)
    if (
        since_ms is not None
        and received_at_ms is not None
        and received_at_ms < since_ms
    ):
        return None

    event_type = "progress"
    tool_name: str | None = None
    tool_state: str | None = None
    summary = text or method or event_name
    normalized_text = text.strip()
    if normalized_text.startswith("🤔"):
        event_type = "assistant_update"
        summary = "Thinking"
    elif normalized_text.startswith("⏳"):
        event_type = "tool_started"
        tool_name = _truncate_tool_name(normalized_text.lstrip("⏳").strip())
        tool_state = "started"
        summary = tool_name or "Tool started"
    elif normalized_text.startswith("✅"):
        event_type = "tool_completed"
        tool_name = _truncate_tool_name(normalized_text.lstrip("✅").strip())
        tool_state = "completed"
        summary = tool_name or "Tool completed"
    elif normalized_text.startswith("❌"):
        event_type = "tool_failed"
        tool_name = _truncate_tool_name(normalized_text.lstrip("❌").strip())
        tool_state = "failed"
        summary = tool_name or "Tool failed"
    elif method == "message.completed":
        event_type = "turn_completed"
        summary = truncate_text(text or "Turn completed", 220)
    else:
        event_type = "assistant_update"
        summary = truncate_text(summary, 220)

    return {
        "event_id": event_id,
        "event_type": event_type,
        "summary": truncate_text(summary, 220),
        "lines": [summary] if summary else [],
        "received_at_ms": received_at_ms,
        "received_at": received_at,
        "tool_name": tool_name,
        "tool_state": tool_state,
    }


async def _list_zeroclaw_tail_events(
    *,
    request: Request,
    workspace_root: str,
    session_id: str,
    turn_id: str,
    resume_after: Optional[int],
    since_ms: Optional[int],
    limit: int,
) -> list[dict[str, Any]]:
    supervisor = getattr(request.app.state, "zeroclaw_supervisor", None)
    list_events = getattr(supervisor, "list_turn_events", None)
    if not callable(list_events):
        return []
    try:
        raw_entries = await list_events(Path(workspace_root), session_id, turn_id)
    except Exception:
        return []
    start_after = int(resume_after or 0)
    serialized: list[dict[str, Any]] = []
    for index, entry in enumerate(raw_entries, start=1):
        if index <= start_after:
            continue
        payload = entry if isinstance(entry, dict) else {}
        serialized_entry = _zeroclaw_tail_event_from_raw(
            event_id=index,
            raw_event=str(payload.get("raw_event") or ""),
            published_at=normalize_optional_text(payload.get("published_at")),
            since_ms=since_ms,
        )
        if serialized_entry is not None:
            serialized.append(serialized_entry)
    if len(serialized) > limit:
        serialized = serialized[-limit:]
    return serialized


def _derive_last_tool(events: list[dict[str, Any]]) -> dict[str, Any] | None:
    last_tool: dict[str, Any] | None = None
    for event in events:
        if not isinstance(event, dict):
            continue
        tool_name = normalize_optional_text(event.get("tool_name"))
        tool_state = normalize_optional_text(event.get("tool_state"))
        if tool_name is None or tool_state is None:
            continue
        if last_tool is None or last_tool.get("name") != tool_name:
            last_tool = {
                "name": tool_name,
                "started_at": None,
                "completed_at": None,
                "status": None,
                "in_flight": False,
            }
        if tool_state == "started":
            last_tool["started_at"] = event.get("received_at")
            last_tool["status"] = "running"
            last_tool["in_flight"] = True
        elif tool_state in {"completed", "failed"}:
            last_tool["completed_at"] = event.get("received_at")
            last_tool["status"] = tool_state
            last_tool["in_flight"] = False
    return last_tool


def _derive_progress_phase(
    *,
    turn_status: str,
    stream_available: bool,
    events: list[dict[str, Any]],
    idle_seconds: Optional[int],
) -> tuple[str, str, str, dict[str, Any] | None]:
    last_tool = _derive_last_tool(events)
    if turn_status == "ok":
        return ("completed", "turn_status", "Turn completed successfully.", last_tool)
    if turn_status == "interrupted":
        return ("interrupted", "turn_status", "Turn was interrupted.", last_tool)
    if turn_status in {"error", "failed"}:
        return ("failed", "turn_status", "Turn failed.", last_tool)

    if last_tool is not None and bool(last_tool.get("in_flight")):
        name = str(last_tool.get("name") or "tool")
        return (
            "waiting_on_tool_call",
            "recent_tool_event",
            f"Waiting on tool '{name}'.",
            last_tool,
        )

    if events:
        last_event_type = str(events[-1].get("event_type") or "").strip().lower()
        if last_event_type in {"assistant_update", "progress"}:
            return (
                "model_running",
                "recent_event",
                "Model is still producing intermediate activity.",
                last_tool,
            )
        if last_event_type in {"tool_completed", "tool_failed"}:
            return (
                "model_running",
                "recent_tool_event",
                "Tool activity finished; waiting for the model to continue or finalize.",
                last_tool,
            )

    idle = int(idle_seconds or 0)
    if not stream_available:
        if idle >= _LIKELY_HUNG_IDLE_SECONDS:
            return (
                "likely_hung",
                "idle_timeout",
                "No recent activity; inspect deeper or retry the interrupt.",
                last_tool,
            )
        if idle >= _NO_STREAM_AVAILABLE_IDLE_SECONDS:
            return (
                "no_stream_available",
                "idle_timeout",
                "Runtime is running but has not emitted streamable progress yet.",
                last_tool,
            )
        return (
            "booting_runtime",
            "runtime_start",
            "Waiting for the runtime to start emitting progress.",
            last_tool,
        )

    if idle >= _LIKELY_HUNG_IDLE_SECONDS:
        return (
            "likely_hung",
            "idle_timeout",
            "No recent activity; inspect deeper or retry the interrupt.",
            last_tool,
        )
    if idle >= _NO_STREAM_AVAILABLE_IDLE_SECONDS:
        return (
            "no_stream_available",
            "idle_timeout",
            "Connected stream has been quiet; the turn may be waiting on backend work.",
            last_tool,
        )
    return (
        "booting_runtime",
        "runtime_start",
        "Waiting for the runtime to emit the first progress event.",
        last_tool,
    )


def _redacted_prompt_preview(value: Any) -> str:
    text = str(value or "")
    if not text:
        return ""
    return truncate_text(redact_text(text), 120)


def _derive_active_turn_diagnostics(
    *,
    snapshot: dict[str, Any],
    turn_record: Optional[dict[str, Any]],
) -> dict[str, Any] | None:
    managed_turn_id = normalize_optional_text(
        snapshot.get("managed_turn_id") or (turn_record or {}).get("managed_turn_id")
    )
    if managed_turn_id is None:
        return None

    events = snapshot.get("events")
    event_list = (
        [event for event in events if isinstance(event, dict)]
        if isinstance(events, list)
        else []
    )
    last_event = event_list[-1] if event_list else {}
    turn_status = str(snapshot.get("turn_status") or "").strip().lower()
    idle_seconds_raw = snapshot.get("idle_seconds")
    idle_seconds = int(idle_seconds_raw) if isinstance(idle_seconds_raw, int) else None
    last_event_at = normalize_optional_text(snapshot.get("last_event_at"))
    stalled = bool(
        turn_status == "running" and idle_seconds is not None and idle_seconds >= 30
    )
    stall_reason = None
    if stalled:
        stall_reason = (
            "no_events_yet"
            if last_event_at is None
            else "no_new_events_since_last_progress"
        )

    return {
        "managed_turn_id": managed_turn_id,
        "request_kind": normalize_optional_text(
            (turn_record or {}).get("request_kind")
        ),
        "model": normalize_optional_text((turn_record or {}).get("model")),
        "reasoning": normalize_optional_text((turn_record or {}).get("reasoning")),
        "prompt_preview": _redacted_prompt_preview((turn_record or {}).get("prompt")),
        "backend_thread_id": normalize_optional_text(snapshot.get("backend_thread_id")),
        "backend_turn_id": normalize_optional_text(
            snapshot.get("backend_turn_id")
            or (turn_record or {}).get("backend_turn_id")
        ),
        "stream_available": bool(snapshot.get("stream_available")),
        "phase": normalize_optional_text(snapshot.get("phase")),
        "guidance": normalize_optional_text(snapshot.get("guidance")),
        "last_event_at": last_event_at,
        "last_event_type": normalize_optional_text(last_event.get("event_type")),
        "last_event_summary": normalize_optional_text(last_event.get("summary")),
        "stalled": stalled,
        "stall_reason": stall_reason,
    }


def _refresh_active_turn_diagnostics(
    snapshot: dict[str, Any],
    *,
    turn_status: Optional[str] = None,
    idle_seconds: Optional[int] = None,
    last_event_at: Optional[str] = None,
    phase: Optional[str] = None,
    guidance: Optional[str] = None,
) -> dict[str, Any] | None:
    diagnostics = snapshot.get("active_turn_diagnostics")
    if not isinstance(diagnostics, dict):
        return None

    updated = dict(diagnostics)
    events = snapshot.get("events")
    event_list = (
        [event for event in events if isinstance(event, dict)]
        if isinstance(events, list)
        else []
    )
    last_event = event_list[-1] if event_list else {}
    resolved_status = (
        str(turn_status or snapshot.get("turn_status") or "").strip().lower()
    )
    resolved_last_event_at = normalize_optional_text(
        last_event_at or snapshot.get("last_event_at")
    )
    if idle_seconds is not None:
        resolved_idle = max(0, int(idle_seconds))
    else:
        resolved_idle = None
        last_event_dt = parse_iso_datetime(resolved_last_event_at)
        if last_event_dt is not None:
            resolved_idle = max(
                0, int((datetime.now(timezone.utc) - last_event_dt).total_seconds())
            )
        else:
            started_dt = parse_iso_datetime(snapshot.get("started_at"))
            if started_dt is not None:
                resolved_idle = max(
                    0, int((datetime.now(timezone.utc) - started_dt).total_seconds())
                )
            elif isinstance(snapshot.get("idle_seconds"), (int, float)):
                resolved_idle = max(0, int(snapshot.get("idle_seconds") or 0))
    stalled = bool(
        resolved_status == "running"
        and resolved_idle is not None
        and resolved_idle >= 30
    )
    updated["phase"] = normalize_optional_text(phase) or normalize_optional_text(
        snapshot.get("phase")
    )
    updated["guidance"] = normalize_optional_text(guidance) or normalize_optional_text(
        snapshot.get("guidance")
    )
    updated["last_event_at"] = resolved_last_event_at
    updated["last_event_type"] = normalize_optional_text(last_event.get("event_type"))
    updated["last_event_summary"] = normalize_optional_text(last_event.get("summary"))
    updated["stalled"] = stalled
    updated["stall_reason"] = (
        "no_events_yet"
        if stalled and resolved_last_event_at is None
        else ("no_new_events_since_last_progress" if stalled else None)
    )
    return updated


def _serialize_tail_event(
    event: dict[str, Any],
    *,
    level: str,
    formatter: AppServerEventFormatter,
    since_ms: Optional[int],
) -> Optional[dict[str, Any]]:
    event_id = int(event.get("id") or 0)
    if event_id <= 0:
        return None
    received_at_ms = int(event.get("received_at") or 0)
    if since_ms is not None and received_at_ms and received_at_ms < since_ms:
        return None
    message = coerce_dict(event.get("message"))
    if _should_suppress_tail_event(message):
        return None
    lines = [
        redact_text(str(line).strip())
        for line in formatter.format_event(message)
        if isinstance(line, str) and line.strip()
    ]
    if not lines:
        fallback = str(message.get("method") or "").strip()
        if fallback:
            lines = [fallback]
    summary = truncate_text(lines[0], 220) if lines else ""
    tool_name, tool_state = _tool_metadata_from_message(message)
    payload: dict[str, Any] = {
        "event_id": event_id,
        "event_type": _tail_event_type_from_message(message),
        "summary": summary,
        "lines": lines[:8],
        "received_at_ms": received_at_ms if received_at_ms > 0 else None,
        "received_at": iso_from_event_ms(received_at_ms),
        "tool_name": tool_name,
        "tool_state": tool_state,
    }
    if level == "debug":
        payload["raw"] = _redact_nested(message)
    return payload


def _event_received_at_iso(event: dict[str, Any]) -> Optional[str]:
    received_at_ms = int(event.get("received_at") or 0)
    if received_at_ms <= 0:
        return None
    return iso_from_event_ms(received_at_ms)


async def _build_managed_thread_tail_snapshot(
    *,
    request: Request,
    service: Any,
    managed_thread_id: str,
    limit: int,
    level: str,
    since_ms: Optional[int],
    resume_after: Optional[int],
) -> dict[str, Any]:
    thread = service.get_thread_target(managed_thread_id)
    if thread is None:
        raise HTTPException(status_code=404, detail="Managed thread not found")
    turn = service.get_running_execution(
        managed_thread_id
    ) or service.get_latest_execution(managed_thread_id)
    if turn is None:
        return {
            "managed_thread_id": managed_thread_id,
            "managed_turn_id": None,
            "agent": thread.agent_id,
            "turn_status": None,
            "lifecycle_events": [],
            "events": [],
            "last_event_id": int(resume_after or 0),
            "elapsed_seconds": None,
            "idle_seconds": None,
            "activity": "idle",
            "stream_available": False,
        }

    managed_turn_id = str(turn.execution_id or "")
    turn_status = str(turn.status or "").strip().lower()
    started_at = normalize_optional_text(turn.started_at)
    finished_at = normalize_optional_text(turn.finished_at)
    started_dt = parse_iso_datetime(started_at)
    finished_dt = parse_iso_datetime(finished_at)
    now_dt = datetime.now(timezone.utc)
    effective_finished = finished_dt or (None if turn_status == "running" else now_dt)
    elapsed_seconds: Optional[int] = None
    if started_dt is not None:
        end_dt = effective_finished or now_dt
        elapsed_seconds = max(0, int((end_dt - started_dt).total_seconds()))

    lifecycle_events = ["turn_started"]
    if turn_status == "ok":
        lifecycle_events.append("turn_completed")
    elif turn_status == "error":
        lifecycle_events.append("turn_failed")
    elif turn_status == "interrupted":
        lifecycle_events.append("turn_interrupted")

    backend_thread_id = normalize_optional_text(thread.backend_thread_id)
    backend_turn_id = normalize_optional_text(turn.backend_id)
    app_server_events = getattr(request.app.state, "app_server_events", None)
    agent_id = str(thread.agent_id or "").strip().lower()
    can_stream_codex = (
        agent_id == "codex"
        and app_server_events is not None
        and bool(backend_thread_id)
        and bool(backend_turn_id)
    )
    can_stream_zeroclaw = (
        agent_id == "zeroclaw"
        and bool(thread.workspace_root)
        and bool(backend_thread_id)
        and bool(backend_turn_id)
        and callable(
            getattr(
                getattr(request.app.state, "zeroclaw_supervisor", None),
                "list_turn_events",
                None,
            )
        )
    )
    formatter = AppServerEventFormatter(redact_enabled=True)
    tail_events: list[dict[str, Any]] = []
    raw_last_activity_at: Optional[str] = None
    if can_stream_codex:
        raw_events = await app_server_events.list_events(
            str(backend_thread_id),
            str(backend_turn_id),
            after_id=int(resume_after or 0),
            limit=limit,
        )
        for event in raw_events:
            activity_at = _event_received_at_iso(event)
            if activity_at:
                raw_last_activity_at = activity_at
            serialized = _serialize_tail_event(
                event,
                level=level,
                formatter=formatter,
                since_ms=since_ms,
            )
            if serialized is not None:
                tail_events.append(serialized)
    elif can_stream_zeroclaw and thread.workspace_root:
        tail_events = await _list_zeroclaw_tail_events(
            request=request,
            workspace_root=str(thread.workspace_root),
            session_id=str(backend_thread_id),
            turn_id=str(backend_turn_id),
            resume_after=resume_after,
            since_ms=since_ms,
            limit=limit,
        )

    last_event_id = int(resume_after or 0)
    last_activity_at: Optional[str] = raw_last_activity_at
    if tail_events:
        last_event_id = int(tail_events[-1].get("event_id") or last_event_id)
        tail_last_activity_at = normalize_optional_text(
            tail_events[-1].get("received_at")
        )
        raw_last_dt = parse_iso_datetime(raw_last_activity_at)
        tail_last_dt = parse_iso_datetime(tail_last_activity_at)
        if raw_last_dt is None:
            last_activity_at = tail_last_activity_at
        elif tail_last_dt is None:
            last_activity_at = raw_last_activity_at
        else:
            last_activity_at = (
                raw_last_activity_at
                if raw_last_dt >= tail_last_dt
                else tail_last_activity_at
            )
    last_event_ms = tail_events[-1].get("received_at_ms") if tail_events else None
    idle_seconds: Optional[int] = None
    if turn_status == "running":
        if last_activity_at:
            last_activity_dt = parse_iso_datetime(last_activity_at)
            if last_activity_dt is not None:
                idle_seconds = max(0, int((now_dt - last_activity_dt).total_seconds()))
        elif isinstance(last_event_ms, (int, float)) and last_event_ms > 0:
            idle_seconds = max(
                0, int((now_dt.timestamp() * 1000 - last_event_ms) / 1000)
            )
        elif started_dt is not None:
            idle_seconds = max(0, int((now_dt - started_dt).total_seconds()))

    activity = "idle"
    if turn_status == "running":
        activity = (
            "stalled" if idle_seconds is not None and idle_seconds >= 30 else "running"
        )
    elif turn_status == "ok":
        activity = "completed"
    elif turn_status == "interrupted":
        activity = "interrupted"
    elif turn_status == "error":
        activity = "failed"

    stream_available = bool(can_stream_codex or can_stream_zeroclaw)
    phase, phase_source, guidance, last_tool = _derive_progress_phase(
        turn_status=turn_status,
        stream_available=stream_available,
        events=tail_events,
        idle_seconds=idle_seconds,
    )
    turn_store = PmaThreadStore(request.app.state.config.root)
    turn_record = turn_store.get_turn(managed_thread_id, managed_turn_id)
    snapshot: dict[str, Any] = {
        "managed_thread_id": managed_thread_id,
        "managed_turn_id": managed_turn_id,
        "agent": thread.agent_id,
        "backend_thread_id": backend_thread_id,
        "backend_turn_id": backend_turn_id,
        "turn_status": turn_status,
        "started_at": started_at,
        "finished_at": finished_at,
        "elapsed_seconds": elapsed_seconds,
        "idle_seconds": idle_seconds,
        "activity": activity,
        "lifecycle_events": lifecycle_events,
        "events": tail_events,
        "last_event_id": last_event_id,
        "last_event_at": tail_events[-1].get("received_at") if tail_events else None,
        "last_activity_at": last_activity_at,
        "stream_available": stream_available,
        "phase": phase,
        "phase_source": phase_source,
        "guidance": guidance,
        "last_tool": last_tool,
    }
    snapshot["active_turn_diagnostics"] = _derive_active_turn_diagnostics(
        snapshot=snapshot,
        turn_record=turn_record,
    )
    return snapshot


def build_managed_thread_tail_routes(
    router: APIRouter,
    get_runtime_state,
) -> None:
    """Build managed-thread status and tail routes."""
    _ = get_runtime_state

    @router.get("/threads/{managed_thread_id}/status")
    async def get_managed_thread_status(
        managed_thread_id: str,
        request: Request,
        limit: int = 20,
        since: Optional[str] = None,
        since_event_id: Optional[int] = None,
        level: str = "info",
    ) -> dict[str, Any]:
        if limit <= 0:
            raise HTTPException(status_code=400, detail="limit must be greater than 0")
        service = build_managed_thread_orchestration_service(request)
        snapshot = await _build_managed_thread_tail_snapshot(
            request=request,
            service=service,
            managed_thread_id=managed_thread_id,
            limit=min(limit, 200),
            level=normalize_tail_level(level),
            since_ms=since_ms_from_duration(since),
            resume_after=resolve_resume_after(request, since_event_id),
        )
        thread = service.get_thread_target(managed_thread_id)
        if thread is None:
            raise HTTPException(status_code=404, detail="Managed thread not found")
        serialized_thread = _serialize_thread_target(thread)
        turn = service.get_running_execution(
            managed_thread_id
        ) or service.get_latest_execution(managed_thread_id)
        queue_store = PmaThreadStore(request.app.state.config.root)
        queued_turns = queue_store.list_pending_turn_queue_items(
            managed_thread_id, limit=min(limit, 50)
        )
        latest_output_excerpt = ""
        if turn is not None:
            latest_output_excerpt = truncate_text(turn.output_text or "", 240)
        turn_status = str(snapshot.get("turn_status") or "")
        return {
            "managed_thread_id": managed_thread_id,
            "thread": serialized_thread,
            "is_alive": bool(
                (serialized_thread.get("lifecycle_status") or "") == "active"
                and turn_status == "running"
            ),
            "status": str(serialized_thread.get("status") or ""),
            "operator_status": str(serialized_thread.get("operator_status") or ""),
            "is_reusable": bool(serialized_thread.get("is_reusable")),
            "status_reason": normalize_optional_text(
                serialized_thread.get("status_reason")
            )
            or "",
            "status_changed_at": normalize_optional_text(
                serialized_thread.get("status_changed_at")
            ),
            "status_terminal": bool(serialized_thread.get("status_terminal")),
            "turn": {
                "managed_turn_id": snapshot.get("managed_turn_id"),
                "status": snapshot.get("turn_status"),
                "activity": snapshot.get("activity"),
                "phase": snapshot.get("phase"),
                "phase_source": snapshot.get("phase_source"),
                "guidance": snapshot.get("guidance"),
                "last_tool": snapshot.get("last_tool"),
                "elapsed_seconds": snapshot.get("elapsed_seconds"),
                "idle_seconds": snapshot.get("idle_seconds"),
                "started_at": snapshot.get("started_at"),
                "finished_at": snapshot.get("finished_at"),
                "lifecycle_events": snapshot.get("lifecycle_events"),
            },
            "queue_depth": service.get_queue_depth(managed_thread_id),
            "queued_turns": [
                {
                    "managed_turn_id": item.get("managed_turn_id"),
                    "request_kind": item.get("request_kind"),
                    "state": item.get("state"),
                    "enqueued_at": item.get("enqueued_at"),
                    "prompt_preview": truncate_text(item.get("prompt") or "", 120),
                }
                for item in queued_turns
            ],
            "recent_progress": snapshot.get("events") or [],
            "latest_output_excerpt": latest_output_excerpt,
            "stream_available": bool(snapshot.get("stream_available")),
            "active_turn_diagnostics": snapshot.get("active_turn_diagnostics"),
        }

    @router.get("/threads/{managed_thread_id}/tail")
    async def get_managed_thread_tail(
        managed_thread_id: str,
        request: Request,
        limit: int = 50,
        since: Optional[str] = None,
        since_event_id: Optional[int] = None,
        level: str = "info",
    ) -> dict[str, Any]:
        if limit <= 0:
            raise HTTPException(status_code=400, detail="limit must be greater than 0")
        service = build_managed_thread_orchestration_service(request)
        return await _build_managed_thread_tail_snapshot(
            request=request,
            service=service,
            managed_thread_id=managed_thread_id,
            limit=min(limit, 200),
            level=normalize_tail_level(level),
            since_ms=since_ms_from_duration(since),
            resume_after=resolve_resume_after(request, since_event_id),
        )

    @router.get("/threads/{managed_thread_id}/tail/events")
    async def stream_managed_thread_tail(
        managed_thread_id: str,
        request: Request,
        limit: int = 50,
        since: Optional[str] = None,
        since_event_id: Optional[int] = None,
        level: str = "info",
    ):
        if limit <= 0:
            raise HTTPException(status_code=400, detail="limit must be greater than 0")
        normalized_level = normalize_tail_level(level)
        since_ms = since_ms_from_duration(since)
        service = build_managed_thread_orchestration_service(request)
        snapshot = await _build_managed_thread_tail_snapshot(
            request=request,
            service=service,
            managed_thread_id=managed_thread_id,
            limit=min(limit, 200),
            level=normalized_level,
            since_ms=since_ms,
            resume_after=resolve_resume_after(request, since_event_id),
        )

        async def _stream() -> Any:
            yield f"event: state\ndata: {json.dumps(snapshot, ensure_ascii=True)}\n\n"
            for event in snapshot.get("events", []):
                if not isinstance(event, dict):
                    continue
                event_id = event.get("event_id")
                event_id_line = (
                    f"id: {event_id}\n"
                    if isinstance(event_id, int) and event_id > 0
                    else ""
                )
                yield (
                    f"event: tail\n"
                    f"{event_id_line}"
                    f"data: {json.dumps(event, ensure_ascii=True)}\n\n"
                )

            if snapshot.get("turn_status") != "running":
                return
            if not snapshot.get("stream_available"):
                while True:
                    await asyncio.sleep(10.0)
                    turn = service.get_execution(
                        managed_thread_id,
                        str(snapshot.get("managed_turn_id") or ""),
                    )
                    status = (
                        str((turn.status if turn is not None else "") or "")
                        .strip()
                        .lower()
                    )
                    if status != "running":
                        yield (
                            "event: state\ndata: "
                            f"{json.dumps({'turn_status': status or 'unknown'}, ensure_ascii=True)}\n\n"
                        )
                        return
                    now = datetime.now(timezone.utc)
                    started_dt = parse_iso_datetime(snapshot.get("started_at"))
                    elapsed = None
                    if started_dt is not None:
                        elapsed = max(0, int((now - started_dt).total_seconds()))
                    phase, phase_source, guidance, last_tool = _derive_progress_phase(
                        turn_status="running",
                        stream_available=False,
                        events=[
                            event
                            for event in snapshot.get("events", [])
                            if isinstance(event, dict)
                        ],
                        idle_seconds=elapsed,
                    )
                    active_turn_diagnostics = _refresh_active_turn_diagnostics(
                        snapshot,
                        turn_status="running",
                        idle_seconds=elapsed,
                        phase=phase,
                        guidance=guidance,
                    )
                    yield (
                        "event: progress\ndata: "
                        f"{json.dumps({'managed_thread_id': managed_thread_id, 'managed_turn_id': snapshot.get('managed_turn_id'), 'turn_status': 'running', 'elapsed_seconds': elapsed, 'phase': phase, 'phase_source': phase_source, 'guidance': guidance, 'last_tool': last_tool, 'active_turn_diagnostics': active_turn_diagnostics}, ensure_ascii=True)}\n\n"
                    )

            if str(snapshot.get("agent") or "").strip().lower() == "zeroclaw":
                thread_target = service.get_thread_target(managed_thread_id)
                workspace_root = (
                    str(thread_target.workspace_root or "")
                    if thread_target is not None
                    else ""
                )
                if (
                    workspace_root
                    and snapshot.get("backend_thread_id")
                    and snapshot.get("backend_turn_id")
                ):
                    last_event_id = int(snapshot.get("last_event_id") or 0)
                    while True:
                        await asyncio.sleep(2.0)
                        incremental_events = await _list_zeroclaw_tail_events(
                            request=request,
                            workspace_root=workspace_root,
                            session_id=str(snapshot.get("backend_thread_id") or ""),
                            turn_id=str(snapshot.get("backend_turn_id") or ""),
                            resume_after=last_event_id,
                            since_ms=since_ms,
                            limit=min(limit, 200),
                        )
                        for serialized in incremental_events:
                            event_id = int(serialized.get("event_id") or 0)
                            if event_id > 0:
                                last_event_id = event_id
                            snapshot_events = snapshot.get("events")
                            if isinstance(snapshot_events, list):
                                snapshot_events.append(serialized)
                            snapshot["last_event_at"] = serialized.get("received_at")
                            yield (
                                "event: tail\n"
                                f"id: {event_id}\n"
                                f"data: {json.dumps(serialized, ensure_ascii=True)}\n\n"
                            )

                        refreshed = await _build_managed_thread_tail_snapshot(
                            request=request,
                            service=service,
                            managed_thread_id=managed_thread_id,
                            limit=min(limit, 200),
                            level=normalized_level,
                            since_ms=since_ms,
                            resume_after=None,
                        )
                        yield (
                            "event: progress\ndata: "
                            f"{json.dumps({'managed_thread_id': managed_thread_id, 'managed_turn_id': refreshed.get('managed_turn_id'), 'turn_status': refreshed.get('turn_status') or 'running', 'elapsed_seconds': refreshed.get('elapsed_seconds'), 'idle_seconds': refreshed.get('idle_seconds'), 'phase': refreshed.get('phase'), 'phase_source': refreshed.get('phase_source'), 'guidance': refreshed.get('guidance'), 'last_tool': refreshed.get('last_tool'), 'active_turn_diagnostics': refreshed.get('active_turn_diagnostics')}, ensure_ascii=True)}\n\n"
                        )
                        if (
                            str(refreshed.get("turn_status") or "").strip().lower()
                            != "running"
                        ):
                            return
                return

            app_server_events = getattr(request.app.state, "app_server_events", None)
            if app_server_events is None:
                return
            backend_thread_id = str(snapshot.get("backend_thread_id") or "")
            backend_turn_id = str(snapshot.get("backend_turn_id") or "")
            if not backend_thread_id or not backend_turn_id:
                return

            formatter = AppServerEventFormatter(redact_enabled=True)
            last_event_id = int(snapshot.get("last_event_id") or 0)
            async for entry in app_server_events.stream_entries(
                backend_thread_id,
                backend_turn_id,
                after_id=last_event_id,
                heartbeat_interval=10.0,
            ):
                if entry is None:
                    turn = service.get_execution(
                        managed_thread_id,
                        str(snapshot.get("managed_turn_id") or ""),
                    )
                    status = (
                        str((turn.status if turn is not None else "") or "")
                        .strip()
                        .lower()
                    )
                    now = datetime.now(timezone.utc)
                    started_dt = parse_iso_datetime(snapshot.get("started_at"))
                    elapsed = None
                    if started_dt is not None:
                        elapsed = max(0, int((now - started_dt).total_seconds()))
                    idle = None
                    activity_at = snapshot.get("last_activity_at") or snapshot.get(
                        "last_event_at"
                    )
                    if activity_at:
                        last_event_dt = parse_iso_datetime(activity_at)
                        if last_event_dt is not None:
                            idle = max(0, int((now - last_event_dt).total_seconds()))
                    phase, phase_source, guidance, last_tool = _derive_progress_phase(
                        turn_status=status or "running",
                        stream_available=True,
                        events=[
                            event
                            for event in snapshot.get("events", [])
                            if isinstance(event, dict)
                        ],
                        idle_seconds=idle,
                    )
                    active_turn_diagnostics = _refresh_active_turn_diagnostics(
                        snapshot,
                        turn_status=status or "running",
                        idle_seconds=idle,
                        phase=phase,
                        guidance=guidance,
                    )
                    yield (
                        "event: progress\ndata: "
                        f"{json.dumps({'managed_thread_id': managed_thread_id, 'managed_turn_id': snapshot.get('managed_turn_id'), 'turn_status': status or 'running', 'elapsed_seconds': elapsed, 'idle_seconds': idle, 'phase': phase, 'phase_source': phase_source, 'guidance': guidance, 'last_tool': last_tool, 'active_turn_diagnostics': active_turn_diagnostics}, ensure_ascii=True)}\n\n"
                    )
                    if status != "running":
                        return
                    continue

                serialized = _serialize_tail_event(
                    entry,
                    level=normalized_level,
                    formatter=formatter,
                    since_ms=since_ms,
                )
                activity_at = _event_received_at_iso(entry)
                if activity_at:
                    snapshot["last_activity_at"] = activity_at
                if serialized is None:
                    continue
                event_id = int(serialized.get("event_id") or 0)
                if event_id > 0:
                    last_event_id = event_id
                snapshot_events = snapshot.get("events")
                if isinstance(snapshot_events, list):
                    snapshot_events.append(serialized)
                snapshot["last_event_at"] = serialized.get("received_at")
                yield (
                    "event: tail\n"
                    f"id: {event_id}\n"
                    f"data: {json.dumps(serialized, ensure_ascii=True)}\n\n"
                )

        return StreamingResponse(
            _stream(),
            media_type="text/event-stream",
            headers=SSE_HEADERS,
        )



"""Shared agent catalog and agent-switch policy for chat surfaces."""

from __future__ import annotations

from dataclasses import dataclass
from types import MappingProxyType
from typing import Literal, Optional

DEFAULT_CHAT_AGENT = "codex"
VALID_CHAT_AGENT_VALUES = ("codex", "opencode")
DEFAULT_CHAT_AGENT_MODELS = MappingProxyType(
    {
        "codex": "gpt-5.3-codex",
        "opencode": "zai-coding-plan/glm-5",
    }
)

AgentModelResetMode = Literal["clear", "agent_default"]


@dataclass(frozen=True)
class ChatAgentSwitchState:
    """Normalized runtime state that should apply after an agent switch."""

    agent: str
    model: Optional[str]
    effort: Optional[str]


def normalize_chat_agent(
    value: object, *, default: Optional[str] = None
) -> Optional[str]:
    if not isinstance(value, str):
        return default
    normalized = value.strip().lower()
    compact = "".join(ch for ch in normalized if ch.isalnum())
    if normalized in VALID_CHAT_AGENT_VALUES:
        return normalized
    if compact in VALID_CHAT_AGENT_VALUES:
        return compact
    return default


def chat_agent_supports_effort(agent: object) -> bool:
    return normalize_chat_agent(agent, default=DEFAULT_CHAT_AGENT) == "codex"


def default_chat_model_for_agent(agent: object) -> Optional[str]:
    normalized = normalize_chat_agent(agent, default=DEFAULT_CHAT_AGENT)
    if normalized is None:
        return None
    return DEFAULT_CHAT_AGENT_MODELS.get(normalized)


def build_agent_switch_state(
    agent: object, *, model_reset: AgentModelResetMode
) -> ChatAgentSwitchState:
    normalized = normalize_chat_agent(agent, default=DEFAULT_CHAT_AGENT)
    if normalized is None:
        normalized = DEFAULT_CHAT_AGENT
    model = (
        default_chat_model_for_agent(normalized)
        if model_reset == "agent_default"
        else None
    )
    return ChatAgentSwitchState(
        agent=normalized,
        model=model,
        effort=None,
    )



from __future__ import annotations

import asyncio
import json
import logging
import re
from dataclasses import dataclass
from pathlib import Path
from typing import Any, AsyncIterator, Optional

from ...core.sse import format_sse, parse_sse_lines
from ..base import AgentHarness
from ..types import (
    AgentId,
    ConversationRef,
    ModelCatalog,
    ModelSpec,
    RuntimeCapability,
    TerminalTurnResult,
    TurnRef,
)
from .constants import DEFAULT_TICKET_MODEL
from .runtime import (
    build_turn_id,
    collect_opencode_output_from_events,
    extract_session_id,
    extract_turn_id,
    map_approval_policy_to_permission,
    split_model_id,
)
from .supervisor import OpenCodeSupervisor

_logger = logging.getLogger(__name__)
_GLOB_META_RE = re.compile(r"[*?\[\]{]")


@dataclass
class _PendingTurnConfig:
    model_payload: Optional[dict[str, str]]
    approval_mode: Optional[str]
    sandbox_policy: Optional[Any]
    question_policy: str
    command_task: Optional[asyncio.Task[Any]] = None


def _path_is_within(root: Path, candidate: Path) -> bool:
    try:
        candidate.relative_to(root)
        return True
    except ValueError:
        return False


def _permission_candidate_path(value: Any, workspace_root: Path) -> Optional[Path]:
    if not isinstance(value, str):
        return None
    text = value.strip()
    if not text:
        return None
    match = _GLOB_META_RE.search(text)
    if match is not None:
        text = text[: match.start()]
    text = text.rstrip("/")
    if not text:
        return None
    path = Path(text)
    if not path.is_absolute():
        path = workspace_root / path
    return path.resolve(strict=False)


def _collect_permission_paths(
    props: dict[str, Any], workspace_root: Path
) -> list[Path]:
    candidates: list[Path] = []

    def _append(value: Any) -> None:
        path = _permission_candidate_path(value, workspace_root)
        if path is not None:
            candidates.append(path)

    for key in ("path", "filepath", "directory"):
        _append(props.get(key))

    patterns = props.get("patterns")
    if isinstance(patterns, list):
        for item in patterns:
            _append(item)

    metadata = props.get("metadata")
    if isinstance(metadata, dict):
        for key in ("path", "filepath", "directory"):
            _append(metadata.get(key))
        nested_patterns = metadata.get("patterns")
        if isinstance(nested_patterns, list):
            for item in nested_patterns:
                _append(item)

    return candidates


def _workspace_permission_decision(
    props: dict[str, Any],
    *,
    workspace_root: Path,
) -> str:
    permission_kind = (
        str(props.get("permission") or props.get("type") or props.get("tool") or "")
        .strip()
        .lower()
    )
    paths = _collect_permission_paths(props, workspace_root)
    if paths:
        if all(_path_is_within(workspace_root, path) for path in paths):
            return "allow"
        return "reject"
    if permission_kind in {"external_directory", "external_file", "external_path"}:
        return "reject"
    return "allow"


def _normalize_message_text(value: Any) -> Optional[str]:
    if isinstance(value, str):
        text = value.strip()
        return text or None
    if isinstance(value, list):
        parts: list[str] = []
        for part in value:
            if isinstance(part, dict):
                part_text = part.get("text")
                if isinstance(part_text, str):
                    parts.append(part_text)
        joined = "".join(parts).strip()
        return joined or None
    if isinstance(value, dict):
        for key in ("text", "message", "content"):
            nested_text = _normalize_message_text(value.get(key))
            if nested_text:
                return nested_text
        return None
    return None


def _extract_delta_text(params: dict[str, Any]) -> Optional[str]:
    for key in ("content", "delta", "text"):
        text = _normalize_message_text(params.get(key))
        if text:
            return text
    properties = params.get("properties")
    if isinstance(properties, dict):
        delta = properties.get("delta")
        if isinstance(delta, dict):
            text = _normalize_message_text(delta)
            if text:
                return text
        part = properties.get("part")
        if isinstance(part, dict) and part.get("type") == "text":
            text = _normalize_message_text(part.get("text"))
            if text:
                return text
    message = params.get("message")
    if isinstance(message, dict):
        return _extract_delta_text(message)
    item = params.get("item")
    if isinstance(item, dict):
        return _extract_delta_text(item)
    return None


def _extract_completed_text(params: dict[str, Any]) -> Optional[str]:
    item = params.get("item")
    if isinstance(item, dict):
        return _normalize_message_text(item.get("content")) or _normalize_message_text(
            item
        )
    result = params.get("result")
    if isinstance(result, dict):
        return _normalize_message_text(result)
    return _normalize_message_text(params)


def _extract_error_text(params: dict[str, Any]) -> Optional[str]:
    error = params.get("error")
    if isinstance(error, dict):
        return _normalize_message_text(error)
    return _normalize_message_text(params.get("message")) or _normalize_message_text(
        params.get("detail")
    )


def _extract_message_info(params: dict[str, Any]) -> dict[str, Any]:
    info = params.get("info")
    if isinstance(info, dict):
        return info
    properties = params.get("properties")
    if isinstance(properties, dict):
        nested = properties.get("info")
        if isinstance(nested, dict):
            return nested
    return {}


def _extract_message_id(params: dict[str, Any]) -> Optional[str]:
    info = _extract_message_info(params)
    for key in ("id", "messageID", "messageId", "message_id"):
        value = info.get(key)
        if isinstance(value, str) and value:
            return value
    for key in ("messageID", "messageId", "message_id"):
        value = params.get(key)
        if isinstance(value, str) and value:
            return value
    return None


def _extract_message_role(params: dict[str, Any]) -> Optional[str]:
    info = _extract_message_info(params)
    role = info.get("role")
    if isinstance(role, str) and role:
        return role
    role = params.get("role")
    if isinstance(role, str) and role:
        return role
    return None


def _normalize_message_phase(value: Any) -> Optional[str]:
    if not isinstance(value, str):
        return None
    normalized = value.strip().lower()
    if normalized in {"commentary", "final_answer"}:
        return normalized
    return None


def _extract_message_phase(params: dict[str, Any]) -> Optional[str]:
    phase = _normalize_message_phase(params.get("phase"))
    if phase:
        return phase
    info = _extract_message_info(params)
    phase = _normalize_message_phase(info.get("phase"))
    if phase:
        return phase
    properties = params.get("properties")
    if isinstance(properties, dict):
        phase = _normalize_message_phase(properties.get("phase"))
        if phase:
            return phase
    message = params.get("message")
    if isinstance(message, dict):
        phase = _normalize_message_phase(message.get("phase"))
        if phase:
            return phase
    item = params.get("item")
    if isinstance(item, dict):
        phase = _normalize_message_phase(item.get("phase"))
        if phase:
            return phase
    return None


def _extract_part_message_id(params: dict[str, Any]) -> Optional[str]:
    properties = params.get("properties")
    part = properties.get("part") if isinstance(properties, dict) else None
    if not isinstance(part, dict):
        return None
    for key in ("messageID", "messageId", "message_id"):
        value = part.get(key)
        if isinstance(value, str) and value:
            return value
    return None


def _unwrap_harness_payload(payload: dict[str, Any]) -> tuple[str, dict[str, Any]]:
    if isinstance(payload.get("message"), dict):
        message = payload["message"]
        method = message.get("method")
        params = message.get("params")
        if isinstance(method, str) and isinstance(params, dict):
            return method, params
    method = payload.get("method")
    params = payload.get("params")
    if isinstance(method, str) and isinstance(params, dict):
        return method, params
    return "", {}


def _collect_terminal_text(payloads: list[dict[str, Any]]) -> tuple[str, list[str]]:
    output_text = ""
    completed_message: Optional[str] = None
    completed_final_message: Optional[str] = None
    commentary_message: Optional[str] = None
    errors: list[str] = []
    message_roles: dict[str, str] = {}
    pending_by_message: dict[str, str] = {}
    pending_no_id = ""
    message_roles_seen = False

    def _merge_stream(current: str, incoming: str) -> str:
        if not incoming:
            return current
        if not current:
            return incoming
        if incoming == current:
            return current
        if len(incoming) > len(current) and incoming.startswith(current):
            return incoming
        max_overlap = min(len(current), max(len(incoming) - 1, 0))
        for overlap in range(max_overlap, 0, -1):
            if current[-overlap:] == incoming[:overlap]:
                return f"{current}{incoming[overlap:]}"
        return f"{current}{incoming}"

    def _append_part_text(message_id: Optional[str], text: str) -> None:
        nonlocal output_text
        nonlocal pending_no_id
        if not text:
            return
        if message_id is None:
            if not message_roles_seen:
                output_text = _merge_stream(output_text, text)
            else:
                pending_no_id = _merge_stream(pending_no_id, text)
            return
        role = message_roles.get(message_id)
        if role == "user":
            return
        if role == "assistant":
            output_text = _merge_stream(output_text, text)
            return
        pending_by_message[message_id] = _merge_stream(
            pending_by_message.get(message_id, ""),
            text,
        )

    def _register_role(message_id: Optional[str], role: Optional[str]) -> None:
        nonlocal output_text
        nonlocal pending_no_id
        nonlocal message_roles_seen
        if not message_id or not role:
            return
        message_roles[message_id] = role
        message_roles_seen = True
        pending = pending_by_message.pop(message_id, "")
        if role == "assistant":
            if pending:
                output_text = _merge_stream(output_text, pending)
            if pending_no_id:
                output_text = _merge_stream(output_text, pending_no_id)
                pending_no_id = ""
            return
        if role == "user":
            pending_no_id = ""

    def _flush_fallback_pending() -> None:
        nonlocal output_text
        if pending_by_message:
            for message_id, pending in list(pending_by_message.items()):
                if message_roles.get(message_id) == "assistant":
                    output_text = _merge_stream(output_text, pending)
            pending_by_message.clear()
        if pending_no_id and not message_roles_seen:
            output_text = _merge_stream(output_text, pending_no_id)

    def _record_completed_message(text: str, phase: Optional[str]) -> None:
        nonlocal commentary_message
        nonlocal completed_final_message
        nonlocal completed_message
        if not text:
            return
        if phase == "final_answer":
            completed_final_message = text
            return
        if phase == "commentary":
            commentary_message = text
            return
        completed_message = text

    for payload in payloads:
        method, params = _unwrap_harness_payload(payload)
        method_lower = method.lower()

        if method in {
            "message.delta",
            "message.updated",
            "message.completed",
            "message.part.updated",
        }:
            if method in {"message.updated", "message.completed"}:
                _register_role(
                    _extract_message_id(params),
                    _extract_message_role(params),
                )
            text = _extract_delta_text(params) or _extract_completed_text(params)
            if text:
                if method == "message.delta":
                    output_text = _merge_stream(output_text, text)
                elif method == "message.part.updated":
                    _append_part_text(_extract_part_message_id(params), text)
                else:
                    role = _extract_message_role(params)
                    if role != "user":
                        _record_completed_message(text, _extract_message_phase(params))
            continue

        if method == "item/agentMessage/delta" or method == "turn/streamDelta":
            text = _extract_delta_text(params)
            if text:
                output_text = _merge_stream(output_text, text)
            continue

        if "outputdelta" in method_lower:
            text = _extract_delta_text(params)
            if text:
                output_text = _merge_stream(output_text, text)
            continue

        if method == "item/completed":
            item = params.get("item")
            if isinstance(item, dict) and item.get("type") == "agentMessage":
                text = _extract_completed_text(params)
                if text:
                    _record_completed_message(text, _extract_message_phase(params))
            continue

        if method in {"turn/error", "error"}:
            error = _extract_error_text(params)
            if error:
                errors.append(error)

    if not completed_message:
        _flush_fallback_pending()
    assistant_text = (
        completed_final_message
        or completed_message
        or output_text
        or commentary_message
        or ""
    ).strip()
    return assistant_text, errors


def _coerce_providers(payload: Any) -> list[dict[str, Any]]:
    if isinstance(payload, dict):
        providers = payload.get("providers")
        if isinstance(providers, list):
            return [entry for entry in providers if isinstance(entry, dict)]
    if isinstance(payload, list):
        return [entry for entry in payload if isinstance(entry, dict)]
    return []


def _iter_provider_models(models_raw: Any) -> list[tuple[str, dict[str, Any]]]:
    models: list[tuple[str, dict[str, Any]]] = []
    if isinstance(models_raw, dict):
        for model_id, model in models_raw.items():
            if isinstance(model_id, str) and model_id:
                if isinstance(model, dict):
                    models.append((model_id, model))
                else:
                    models.append((model_id, {"id": model_id}))
        return models
    if isinstance(models_raw, list):
        for entry in models_raw:
            if isinstance(entry, dict):
                model_id = entry.get("id") or entry.get("modelID")
                if isinstance(model_id, str) and model_id:
                    models.append((model_id, entry))
            elif isinstance(entry, str) and entry:
                models.append((entry, {"id": entry}))
    return models


class OpenCodeHarness(AgentHarness):
    agent_id: AgentId = AgentId("opencode")
    display_name = "OpenCode"
    capabilities = frozenset(
        [
            RuntimeCapability("durable_threads"),
            RuntimeCapability("message_turns"),
            RuntimeCapability("interrupt"),
            RuntimeCapability("active_thread_discovery"),
            RuntimeCapability("review"),
            RuntimeCapability("model_listing"),
            RuntimeCapability("event_streaming"),
        ]
    )

    def __init__(self, supervisor: OpenCodeSupervisor) -> None:
        self._supervisor = supervisor
        self._pending_turns: dict[tuple[str, str], _PendingTurnConfig] = {}

    async def ensure_ready(self, workspace_root: Path) -> None:
        await self._supervisor.get_client(workspace_root)

    async def model_catalog(self, workspace_root: Path) -> ModelCatalog:
        client = await self._supervisor.get_client(workspace_root)
        payload = await client.providers(directory=str(workspace_root))
        providers = _coerce_providers(payload)
        models: list[ModelSpec] = []
        default_model = ""
        if isinstance(payload, dict):
            raw_default = payload.get("default")
            if isinstance(raw_default, dict):
                for provider in providers:
                    provider_id = provider.get("id") or provider.get("providerID")
                    if (
                        isinstance(provider_id, str)
                        and provider_id
                        and provider_id in raw_default
                    ):
                        default_model_id = raw_default[provider_id]
                        if isinstance(default_model_id, str) and default_model_id:
                            default_model = f"{provider_id}/{default_model_id}"
                            break
        for provider in providers:
            provider_id = provider.get("id") or provider.get("providerID")
            if not isinstance(provider_id, str) or not provider_id:
                continue
            models_raw = provider.get("models")
            for model_id, model in _iter_provider_models(models_raw):
                name = model.get("name") or model.get("id") or model_id
                display_name = name if isinstance(name, str) and name else model_id
                capabilities = model.get("capabilities")
                supports_reasoning = False
                if isinstance(capabilities, dict):
                    supports_reasoning = bool(capabilities.get("reasoning"))
                variants = model.get("variants")
                reasoning_options: list[str] = []
                if isinstance(variants, dict):
                    reasoning_options = [
                        key for key in variants.keys() if isinstance(key, str)
                    ]
                    if reasoning_options:
                        supports_reasoning = True
                models.append(
                    ModelSpec(
                        id=f"{provider_id}/{model_id}",
                        display_name=display_name,
                        supports_reasoning=supports_reasoning,
                        reasoning_options=reasoning_options,
                    )
                )
        if not default_model and models:
            default_model = models[0].id
        return ModelCatalog(default_model=default_model, models=models)

    async def new_conversation(
        self, workspace_root: Path, title: Optional[str] = None
    ) -> ConversationRef:
        client = await self._supervisor.get_client(workspace_root)
        result = await client.create_session(
            title=title,
            directory=str(workspace_root),
        )
        session_id = extract_session_id(result) or result.get("id")
        if not isinstance(session_id, str) or not session_id:
            raise ValueError("OpenCode did not return a session id")
        return ConversationRef(agent=AgentId("opencode"), id=session_id)

    async def list_conversations(self, workspace_root: Path) -> list[ConversationRef]:
        client = await self._supervisor.get_client(workspace_root)
        result = await client.list_sessions(directory=str(workspace_root))
        sessions: list[dict[str, Any]] = []
        if isinstance(result, dict):
            data = result.get("data")
            if isinstance(data, list):
                sessions = [entry for entry in data if isinstance(entry, dict)]
        elif isinstance(result, list):
            sessions = [entry for entry in result if isinstance(entry, dict)]
        conversations: list[ConversationRef] = []
        for entry in sessions:
            session_id = extract_session_id(entry) or entry.get("id")
            if isinstance(session_id, str) and session_id:
                conversations.append(
                    ConversationRef(agent=AgentId("opencode"), id=session_id)
                )
        return conversations

    async def resume_conversation(
        self, workspace_root: Path, conversation_id: str
    ) -> ConversationRef:
        client = await self._supervisor.get_client(workspace_root)
        try:
            result = await client.get_session(conversation_id)
        except Exception:
            result = {}
        session_id = extract_session_id(result) or conversation_id
        return ConversationRef(agent=AgentId("opencode"), id=session_id)

    async def start_turn(
        self,
        workspace_root: Path,
        conversation_id: str,
        prompt: str,
        model: Optional[str],
        reasoning: Optional[str],
        *,
        approval_mode: Optional[str],
        sandbox_policy: Optional[Any],
        input_items: Optional[list[dict[str, Any]]] = None,
    ) -> TurnRef:
        _ = input_items
        client = await self._supervisor.get_client(workspace_root)
        if model is None:
            model = DEFAULT_TICKET_MODEL
        model_payload = split_model_id(model)
        turn_id = build_turn_id(conversation_id)
        await client.prompt_async(
            conversation_id,
            message=prompt,
            model=model_payload,
            variant=reasoning,
        )
        self._pending_turns[(conversation_id, turn_id)] = _PendingTurnConfig(
            model_payload=model_payload,
            approval_mode=approval_mode,
            sandbox_policy=sandbox_policy,
            question_policy=(
                "reject"
                if approval_mode is not None or sandbox_policy is not None
                else "ignore"
            ),
        )
        return TurnRef(
            conversation_id=conversation_id,
            turn_id=turn_id,
        )

    async def start_review(
        self,
        workspace_root: Path,
        conversation_id: str,
        prompt: str,
        model: Optional[str],
        reasoning: Optional[str],
        *,
        approval_mode: Optional[str],
        sandbox_policy: Optional[Any],
    ) -> TurnRef:
        client = await self._supervisor.get_client(workspace_root)
        if model is None:
            model = DEFAULT_TICKET_MODEL
        arguments = prompt if prompt else ""
        turn_id = build_turn_id(conversation_id)

        async def _send_review() -> None:
            result = await client.send_command(
                conversation_id,
                command="review",
                arguments=arguments,
                model=model,
            )
            started_turn_id = extract_turn_id(conversation_id, result)
            if started_turn_id:
                _logger.debug("OpenCode review started: %s", started_turn_id)

        self._pending_turns[(conversation_id, turn_id)] = _PendingTurnConfig(
            model_payload=split_model_id(model),
            approval_mode=approval_mode,
            sandbox_policy=sandbox_policy,
            question_policy=(
                "reject"
                if approval_mode is not None or sandbox_policy is not None
                else "ignore"
            ),
            command_task=asyncio.create_task(_send_review()),
        )
        return TurnRef(conversation_id=conversation_id, turn_id=turn_id)

    async def interrupt(
        self, workspace_root: Path, conversation_id: str, turn_id: Optional[str]
    ) -> None:
        client = await self._supervisor.get_client(workspace_root)
        try:
            await client.abort(conversation_id)
        except Exception as exc:
            _logger.debug(
                "Failed to abort OpenCode session %s: %s", conversation_id, exc
            )

    async def stream_events(
        self, workspace_root: Path, conversation_id: str, turn_id: str
    ) -> AsyncIterator[str]:
        client = await self._supervisor.get_client(workspace_root)
        async for event in client.stream_events(directory=str(workspace_root)):
            payload = event.data
            try:
                parsed = json.loads(payload) if payload else {}
            except json.JSONDecodeError:
                parsed = {"raw": payload}
            session_id = extract_session_id(parsed)
            status_type = None
            if event.event == "session.status" and isinstance(parsed, dict):
                properties = parsed.get("properties")
                if isinstance(properties, dict):
                    status = properties.get("status") or {}
                else:
                    status = parsed.get("status") or {}
                if isinstance(status, dict):
                    status_type = status.get("type") or status.get("status")
            if (
                event.event == "session.idle"
                or (
                    event.event == "session.status"
                    and isinstance(status_type, str)
                    and status_type.lower() == "idle"
                )
            ) and session_id == conversation_id:
                break
            if session_id and session_id != conversation_id:
                continue
            if not session_id:
                continue
            wrapped = {"message": {"method": event.event, "params": parsed}}
            yield format_sse("app-server", wrapped)

    async def wait_for_turn(
        self,
        workspace_root: Path,
        conversation_id: str,
        turn_id: Optional[str],
        *,
        timeout: Optional[float] = None,
    ) -> TerminalTurnResult:
        pending = self._pending_turns.get((conversation_id, turn_id or ""))
        client = await self._supervisor.get_client(workspace_root)
        workspace_root = workspace_root.resolve()

        async def _collect() -> TerminalTurnResult:
            if pending is None:
                payloads: list[dict[str, Any]] = []

                async def _iter_lines(raw_event_text: str) -> AsyncIterator[str]:
                    for line in raw_event_text.splitlines():
                        yield line
                    yield ""

                async for raw_event in self.stream_events(
                    workspace_root,
                    conversation_id,
                    turn_id or "",
                ):
                    async for sse_event in parse_sse_lines(_iter_lines(str(raw_event))):
                        try:
                            payload = (
                                json.loads(sse_event.data) if sse_event.data else {}
                            )
                        except json.JSONDecodeError:
                            payload = {}
                        if isinstance(payload, dict):
                            payloads.append(payload)

                assistant_text, errors = _collect_terminal_text(payloads)
                return TerminalTurnResult(
                    status="error" if errors else "ok",
                    assistant_text=assistant_text,
                    errors=errors,
                    raw_events=payloads,
                )

            raw_events: list[dict[str, Any]] = []

            async def _event_stream() -> AsyncIterator[Any]:
                async for event in client.stream_events(directory=str(workspace_root)):
                    payload = event.data
                    try:
                        parsed = json.loads(payload) if payload else {}
                    except json.JSONDecodeError:
                        parsed = {"raw": payload}
                    if isinstance(parsed, dict):
                        wrapped = {"message": {"method": event.event, "params": parsed}}
                        raw_events.append(wrapped)
                    yield event

            async def _fetch_session() -> Any:
                statuses = await client.session_status(directory=str(workspace_root))
                if isinstance(statuses, dict):
                    session_status = statuses.get(conversation_id)
                    if session_status is None:
                        return {"status": {"type": "idle"}}
                    if isinstance(session_status, dict):
                        return {"status": session_status}
                    if isinstance(session_status, str):
                        return {"status": session_status}
                return {"status": {}}

            async def _fetch_providers() -> Any:
                return await client.providers(directory=str(workspace_root))

            async def _respond_permission(request_id: str, reply: str) -> None:
                await client.respond_permission(request_id=request_id, reply=reply)

            async def _reply_question(
                request_id: str, answers: list[list[str]]
            ) -> None:
                await client.reply_question(request_id, answers=answers)

            async def _reject_question(request_id: str) -> None:
                await client.reject_question(request_id)

            permission_policy = map_approval_policy_to_permission(
                pending.approval_mode if pending is not None else None,
                default="allow",
            )
            permission_handler = None
            if permission_policy == "ask":

                async def _permission_handler(
                    _request_id: str, props: dict[str, Any]
                ) -> str:
                    return _workspace_permission_decision(
                        props,
                        workspace_root=workspace_root,
                    )

                permission_handler = _permission_handler

            collect_task = asyncio.create_task(
                collect_opencode_output_from_events(
                    None,
                    session_id=conversation_id,
                    model_payload=(
                        pending.model_payload if pending is not None else None
                    ),
                    permission_policy=permission_policy,
                    permission_handler=permission_handler,
                    question_policy=(
                        pending.question_policy if pending is not None else "ignore"
                    ),
                    respond_permission=_respond_permission,
                    reply_question=_reply_question,
                    reject_question=_reject_question,
                    event_stream_factory=_event_stream,
                    session_fetcher=_fetch_session,
                    provider_fetcher=_fetch_providers,
                    stall_timeout_seconds=self._supervisor.session_stall_timeout_seconds,
                )
            )
            command_task = pending.command_task if pending is not None else None
            if command_task is None:
                output = await collect_task
            else:
                try:
                    done, _pending = await asyncio.wait(
                        {collect_task, command_task},
                        return_when=asyncio.FIRST_EXCEPTION,
                    )
                    if command_task in done:
                        exc = command_task.exception()
                        if exc is not None:
                            collect_task.cancel()
                            try:
                                await collect_task
                            except asyncio.CancelledError:
                                pass
                            raise exc
                    output = await collect_task
                    await command_task
                except Exception as exc:
                    return TerminalTurnResult(
                        status="error",
                        assistant_text="",
                        errors=[str(exc)],
                        raw_events=raw_events,
                    )

            errors = [output.error] if output.error else []
            return TerminalTurnResult(
                status="error" if errors else "ok",
                assistant_text=output.text,
                errors=errors,
                raw_events=raw_events,
            )

        try:
            if timeout is None:
                return await _collect()
            return await asyncio.wait_for(_collect(), timeout=timeout)
        finally:
            if turn_id is not None:
                self._pending_turns.pop((conversation_id, turn_id), None)


__all__ = ["OpenCodeHarness"]
